import io
import json
from relevanceai.apps.report_app.blocks import ReportBlocks


class AltairReportBlock(ReportBlocks):
    """
    Adding a Altair block to Report
    """

    def altair(
        self,
        fig,
        title: str = "",
        static: bool = False,
        width: int = 600,
        height: int = 300,
        add: bool = True,
        width_percentage: int = 50,
        **kwargs,
    ):
        try:
            import altair
        except ImportError:
            raise ImportError(
                ".altair requires altair to be installed, install with 'pip install -U altair'."
            )
        if static:
            try:
                import altair_saver

                fp = "_test_.png"
                fig.to_save(fp, format="png")
                self.image(fp, title=title, width_percentage=width_percentage, add=add)
            except ImportError:
                raise ImportError(
                    ".altair 'static=True' requires altair_saver to be installed, install with 'pip install -U altair_saver'."
                )
        else:
            block = {
                "type": "appBlock",
                # "attrs" : {"id": str(uuid.uuid4())},
                "content": [
                    {
                        "attrs": {
                            "height": "auto",
                            "title": title,
                            "width": f"{width_percentage}%",
                            "spec": json.loads(fig.to_json()),
                        },
                        "type": "vegaChart",
                    }
                ],
            }
            if add:
                self.contents.append(block)
            return block

    altair_chart = altair


from abc import ABC, abstractmethod

from typing import Any, List, Union

from relevanceai.utils import DocUtils


class ScalerModelBase(ABC, DocUtils):
    model_name: str
    alias: Union[str, None]

    @staticmethod
    def import_from_string(name):
        """It takes a string, splits it on the period, and then imports the module

        Parameters
        ----------
        name
            The name of the class to import.

        Returns
        -------
            The module object.

        """
        components = name.split(".")
        mod = __import__(components[0])
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod

    def vector_name(self, field: str) -> str:
        """If the alias is a string, then return the alias if it contains the string "_vector_", otherwise
        return the alias with "_vector_" appended to it.

        If the alias is not a string, then return "concat_model_name_vector_" if there are more than one
        fields, otherwise return "field_model_name_vector_"

        Parameters
        ----------
        fields : List[str]
            List[str]

        Returns
        -------
            The name of the vector.

        """
        if isinstance(self.alias, str):
            if "_vector_" in self.alias:
                return self.alias
            else:
                return f"{self.alias}_vector_"
        else:

            if "_vector_" in field:
                field_name = field.split("_vector_")[0]

            return f"{field_name}_{self.model_name}_vector_"

    def fit(self, *args, **kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def fit_transform(self, *args, **kwargs) -> Any:
        raise NotImplementedError



"""
Build a co-occurrence network based on the documents in your dataset
"""
import sys

from collections import defaultdict
import numpy as np
from sklearn.cluster import AgglomerativeClustering
from relevanceai.operations_new.transform_base import TransformBase
from relevanceai.operations_new.processing.text.clean.transform import CleanTextTransform
from relevanceai.constants.stopwords import STOPWORDS


def preprocess(data, stopwords_list=[]):
    # stopwords_list = STOPWORDS # stopwords is a list of strings.
    # stopwords_list += stopwords
    cleaner = CleanTextTransform(text_fields=[], output_fields=[], lower=True,
                                 lemmatize=True, remove_stopwords=STOPWORDS + stopwords_list)
    res = []
    for doc in data:
        res.append(cleaner.clean_text(doc).split(' '))
    return res


class WordDictionary():
    def __init__(self, docs):
        self.id2word = []
        self.word2id = dict()
        self.id2dfs = dict()
        self.id2docs = defaultdict(list)
        self.doc2ids = defaultdict(list)
        for doc_id, doc in enumerate(docs):
            words = set(doc)
            for word in words:
                if word not in self.word2id:
                    self.word2id[word] = len(self.id2word)
                    self.id2word.append(word)
                    self.id2dfs[self.word2id[word]] = 1
                else:
                    self.id2dfs[self.word2id[word]] = self.id2dfs.get(self.word2id[word]) + 1
                self.doc2ids[doc_id].append(self.word2id[word])
                self.id2docs[self.word2id[word]].append(doc_id)

    def get_df_table(self):
        return self.id2dfs

    def get_ids(self, doc_id):
        return self.doc2ids[doc_id]

    def get_docs(self, word: str):
        return self.id2docs[self.word2id[word]]

    def get_word(self, id):
        return self.id2word[id]

    def get_id(self, word: str):
        return self.word2id[word]

    # return updated df_table based on only the documents contains word
    def update_df_table(self, word: str):
        df_table = dict()
        for doc_id in self.get_docs(word):
            for word_id in self.get_ids(doc_id):
                if word_id not in df_table:
                    df_table[word_id] = 1
                else:
                    df_table[word_id] = df_table[word_id] + 1
        return df_table


class CoOccurNetTransform(TransformBase):
    def __init__(
            self,
            max_number_of_clusters=15,
            min_number_of_clusters=3,
            number_of_concepts=100,
            **kwargs,
    ):
        """
        Parameters
        -------------

        max_number_of_clusters: int

        min_number_of_clusters: int

        number_of_concepts: int
        """

        self.max_number_of_clusters = max_number_of_clusters
        self.min_number_of_clusters = min_number_of_clusters
        self.number_of_concepts = number_of_concepts
        for k, v in kwargs.items():
            setattr(self, k, v)

    def find_max_vertex(self, visited, weights):
        # Stores the index of max-weight vertex
        # from set of unvisited vertices
        index = -1

        # Stores the maximum weight from
        # the set of unvisited vertices
        max_weight = -sys.maxsize

        # Iterate over all possible
        # Nodes of a graph
        for i in range(self.number_of_concepts):
            # If the current Node is unvisited
            # and weight of current vertex is
            # greater than max_weight
            if visited[i] is False and weights[i] > max_weight:
                # Update max_weight
                max_weight = weights[i]
                # Update index
                index = i
        return index

    # Function to find the maximum spanning tree
    def maximum_spanning_tree(self, graph):
        # visited[i]:Check if vertex i
        # is visited or not
        visited = [True] * self.number_of_concepts

        # weights[i]: Stores maximum weight of
        # graph to connect an edge with i
        weights = [0] * self.number_of_concepts

        # parent[i]: Stores the parent Node
        # of vertex i
        parent = [0] * self.number_of_concepts

        # Initialize weights as -INFINITE,
        # and visited of a Node as False
        for i in range(self.number_of_concepts):
            visited[i] = False
            weights[i] = -sys.maxsize

        # Include 1st vertex in
        # maximum spanning tree
        weights[0] = sys.maxsize
        parent[0] = -1

        # Search for other (V-1) vertices
        # and build a tree
        for i in range(self.number_of_concepts - 1):

            # Stores index of max-weight vertex
            # from a set of unvisited vertex
            max_vertex_index = self.find_max_vertex(visited, weights)

            # Mark that vertex as visited
            visited[max_vertex_index] = True

            # Update adjacent vertices of
            # the current visited vertex
            for j in range(self.number_of_concepts):

                # If there is an edge between j
                # and current visited vertex and
                # also j is unvisited vertex
                if graph[j][max_vertex_index] != 0 and visited[j] == False:

                    # If graph[v][x] is
                    # greater than weight[v]
                    if graph[j][max_vertex_index] > weights[j]:
                        # Update weights[j]
                        weights[j] = graph[j][max_vertex_index]

                        # Update parent[j]
                        parent[j] = max_vertex_index
        return parent

    def concurrence_matrix(self, top_ids, texts, word_dict):
        word_count_mat = []
        for i, text in enumerate(texts):
            row = [0] * self.number_of_concepts
            for word in text:
                id = word_dict.word2id[word]
                if id in top_ids:
                    row[top_ids.index(id)] = 1
            word_count_mat.append(row)
        word_count_mat = np.array(word_count_mat)

        return np.dot(word_count_mat.transpose(), word_count_mat)

    def get_clusters_labels(self, mat):
        labels = np.zeros((self.number_of_concepts, self.number_of_concepts), dtype=int)

        for i in range(self.number_of_concepts):
            labels[i, 0] = i
        for i, merged in enumerate(
                AgglomerativeClustering(affinity='precomputed', n_clusters=i, linkage='complete').fit(mat).children_):
            new_id = i + self.number_of_concepts
            for j in range(self.number_of_concepts):
                if labels[j, i] in merged:
                    labels[j, i + 1] = new_id
                else:
                    labels[j, i + 1] = labels[j, i]

        for i in range(self.number_of_concepts):
            unique_values = np.unique(labels[:, i])
            label2label = dict(zip(unique_values, range(len(unique_values))))
            for j, v in enumerate(labels[:, i]):
                labels[j, i] = label2label[v]

        return labels

    def transform(self, documents, text_field='content', stopwords_list=[], center_word=None):
        # For each document, update the field
        docs = [d[text_field] for d in documents]
        cleaned_texts = preprocess(docs, stopwords_list)

        # Create Dictionary
        word_dict = WordDictionary(cleaned_texts)

        # Create Corpus
        if center_word is None:
            texts = cleaned_texts
            df_table = word_dict.get_df_table()
        else:
            doc_id_list = word_dict.get_docs(center_word)
            texts = [cleaned_texts[i] for i in doc_id_list]
            df_table = word_dict.update_df_table(center_word)

        top_ids = sorted(df_table, key=df_table.get, reverse=True)[:self.number_of_concepts]
        co_occur_mat = self.concurrence_matrix(top_ids, texts, word_dict)
        mst = self.maximum_spanning_tree(co_occur_mat)

        vertexes = []
        for i, id in enumerate(top_ids):
            v = {'word': word_dict.id2word[id], 'rank': i, 'count': df_table.get(id)}
            vertexes.append(v)

        labels = self.get_clusters_labels(co_occur_mat[0][0] - co_occur_mat)
        for i in range(self.min_number_of_clusters, self.max_number_of_clusters + 1):
            for j in range(self.number_of_concepts):
                vertexes[j]['label_{}'.format(i)] = labels[j, self.number_of_concepts - i]

        edges = []
        for sour, dest in enumerate(mst[1:]):
            edges.append((word_dict.id2word[top_ids[dest]], word_dict.id2word[top_ids[sour+1]]))

        return vertexes, edges

    @property
    def name(self):
        return "co_occurrence_network"



from copy import deepcopy

from typing import List

from relevanceai.operations_new.transform_base import TransformBase
from relevanceai.utils import MissingPackageError


class SentenceSplittingTransform(TransformBase):
    def __init__(
        self,
        text_fields: List[str],
        language: str = "en",
        inplace: bool = True,
        output_field: str = "_splittextchunk_",
    ):
        self.text_fields = text_fields
        self.language = language
        self.inplace = inplace
        self.output_field = output_field

    @property
    def name(self):
        return "sentence_splitting"

    def split_text(
        self,
        text,
    ):
        """The function takes a text and a language as input and returns a list of sentences

        Parameters
        ----------
        text
            The text to split into sentences.
        language : str, optional
            The language of the text.

        Returns
        -------
            A list of sentences.

        """
        try:
            from sentence_splitter import split_text_into_sentences
        except ModuleNotFoundError:
            raise MissingPackageError("sentence-splitter")

        return split_text_into_sentences(text=text, language=self.language)

    def split_text_document(
        self,
        document,
    ):
        """It takes a document, splits it into chunks, and then returns a list of documents, each
        containing a single chunk

        Parameters
        ----------
        text_fields
            The fields in the document that contain the text to be split.
        document
            The document to be split.
        output_field : str, optional
            The name of the field that will contain the split text.

        Returns
        -------
            A list of dictionaries.

        """
        for text_field in self.text_fields:
            text = self.get_fields(text_field, document)
            split_text = self.split_text(text)

            # Format the split text into documents
            split_text_value = [{text_field: s} for s in split_text if s.strip() != ""]
            self.set_field(
                self.output_field,
                document,
                split_text_value,
            )

        return document

    def transform(
        self,
        documents,
    ):
        """It takes a list of documents, and for each document, it splits the text into chunks and adds the
        chunks to the document

        Parameters
        ----------
        text_fields
            a list of fields in the document that contain text to be split
        documents
            list of documents to be split
        inplace : bool, optional
            bool = True
        output_field : str, optional
            str = "_splittextchunk_"

        Returns
        -------
            A list of documents

        """
        if not self.inplace:
            documents = deepcopy(documents)

        # TODO; switch to something faster than list comprehension
        [
            self.split_text_document(
                document=document,
                text_fields=self.text_fields,
                output_field=self.output_field,
            )
            for document in documents
        ]
        return documents



from typing import Dict, List, Optional, Union

from relevanceai.client.helpers import Credentials
from relevanceai.dataset.series import Series
from relevanceai.dataset.apps import LaunchApps
from relevanceai.operations import Operations
from relevanceai.operations_new import Operations as OperationsNew
from relevanceai.utils.decorators.analytics import track
from relevanceai.constants import GLOBAL_DATASETS


class Dataset(OperationsNew, Operations, LaunchApps):
    @track
    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
        fields: Optional[list] = None,
        image_fields: Optional[List[str]] = None,
        audio_fields: Optional[List[str]] = None,
        highlight_fields: Optional[Dict[str, List]] = None,
        text_fields: Optional[List[str]] = None,
        **kwargs,
    ):
        self.credentials = credentials
        self.fields = [] if fields is None else fields
        self.dataset_id = dataset_id
        self.image_fields = [] if image_fields is None else image_fields
        self.audio_fields = [] if audio_fields is None else audio_fields
        self.highlight_fields = {} if highlight_fields is None else highlight_fields
        self.text_fields = [] if text_fields is None else text_fields

        super().__init__(
            credentials=credentials,
            dataset_id=self.dataset_id,
            fields=fields,
            image_fields=image_fields,
            audio_fields=audio_fields,
            highlight_fields=highlight_fields,
            text_fields=text_fields,
            **kwargs,
        )
        # add global datasets
        if self.dataset_id in GLOBAL_DATASETS:
            # avoid re-inserting if it already exists
            if self.shape[0] == 0:
                from relevanceai.utils.datasets import mock_documents
                from relevanceai.utils.decorators.analytics import fire_and_forget

                @fire_and_forget
                def add_mock_dataset():
                    self.upsert_documents(mock_documents(100))

                add_mock_dataset()
        self.is_empty()

    def is_empty(self):
        """Check if a dataset is empty."""
        try:
            if self.shape[0] == 0:
                try:
                    print("⚠️ Your dataset has no documents. Make sure to insert some!")
                except:
                    pass
        except:
            pass

    @track
    def __getitem__(self, field: Union[List[str], str]):
        """
        Returns a Series Object that selects a particular field within a dataset
        Parameters
        ----------
        field: Union[List, str]
            The particular field within the dataset
        Returns
        -------
        Tuple
            (N, C)
        Example
        ---------------
        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            field = "sample_field"
            series = df[field]
        """
        if isinstance(field, str):
            return Series(
                dataset=self,
                field=field,
                image_fields=self.image_fields,
                audio_fields=self.audio_fields,
                highlight_fields=self.highlight_fields,
                text_fields=self.text_fields,
            )
        elif isinstance(field, list):
            return Dataset(
                credentials=self.credentials,
                dataset_id=self.dataset_id,
                fields=field,
                image_fields=self.image_fields,
                audio_fields=self.audio_fields,
                highlight_fields=self.highlight_fields,
                text_fields=self.text_fields,
            )
        else:
            raise TypeError("Field needs to be a list or a string.")

    def set_dtypes(self, mapping: dict):
        unstruc_types = ["_numeric_", "_category_", "_text_", "_image_"]
        for unstruc_type in unstruc_types:
            if unstruc_type not in mapping:
                mapping[unstruc_type] = []

        self.datasets.post_metadata(
            self.dataset_id,
            metadata=mapping,
        )

    def get_dtypes(self):
        metadata = self.datasets.metadata(
            self.dataset_id,
        )["results"]
        metadata = {
            key: value
            for key, value in metadata.items()
            if key.startswith("_") and key.endswith("_")
        }
        return metadata

    def label_clusters(self, cluster_labels: dict, alias: str, vector_fields: list):
        """
        Label your clusters programatiically

        Example
        ----------

        .. code-block::

            ds.label_clusters(
                {"cluster_1" : "nice reviews"},
                alias=...,
                vector_fields=...
            )

        """
        metadata = self.metadata
        metadata_doc = metadata.to_dict()

        if "cluster_metadata" not in metadata_doc:
            metadata_doc["cluster_metadata"] = {"labels": {}}

        cluster_field = "_cluster_." + ".".join(vector_fields) + "." + alias

        if cluster_field not in metadata_doc["cluster_metadata"]["labels"]:
            labels = metadata_doc["cluster_metadata"]["labels"]
            labels.update({cluster_field: {"labels": cluster_labels}})
        else:
            metadata_doc["cluster_metadata"]["labels"][cluster_field][
                "labels"
            ] = cluster_labels
        return self.upsert_metadata(metadata_doc)

    def _return_sort_in_metrics(self, metric_name, metrics):
        if metric_name in [m["name"] for m in metrics]:
            return metric_name
        elif metric_name in [m["field"] for m in metrics]:
            for m in metrics:
                if metric_name == m["field"]:
                    return m["name"]
        else:
            raise KeyError(f"'{metric_name}' is not found in 'metrics'")

    def _clean_metrics(self, metrics, name_prefix="Average "):
        main_metrics = []
        metric_fields = []
        metric_names = []
        for m in metrics:
            if isinstance(m, str):
                main_metrics.append(
                    {"agg": "avg", "field": m, "name": f"{name_prefix}{m}"}
                )
                metric_fields.append(m)
                metric_names.append(f"{name_prefix}{m}")
            else:
                main_metrics.append(m)
                metric_fields.append(m["field"])
                metric_names.append(m["name"])
        return main_metrics, metric_fields, metric_names

    def _clean_groupby(self, groupby):
        main_groupby = []
        groupby_fields = []
        for m in groupby:
            if isinstance(m, str):
                if self.schema[m] == "text":
                    main_groupby.append({"agg": "category", "field": m, "name": f"{m}"})
                elif self.schema[m] == "numeric":
                    main_groupby.append({"agg": "numeric", "field": m, "name": f"{m}"})
                elif self.schema[m] == "date":
                    main_groupby.append(
                        {
                            "agg": "category",
                            "field": m,
                            "name": f"{m}",
                            "date_interval": "monthly",
                        }
                    )
                groupby_fields.append(m)
            else:
                main_groupby.append(m)
                groupby_fields.append(m["field"])
        return main_groupby, groupby_fields

    def _auto_detect_vector_fields(
        self, fields, vector_fields="auto", field_type="text"
    ):
        if vector_fields == "auto":
            vector_fields = []
            print(
                f'Detected "{field_type}_vector_fields" is set as "auto", will try to determine "{field_type}_vector_fields" from "{field_type}_fields"'
            )
            for field, field_type in self.schema.items():
                if isinstance(field_type, dict):
                    for f in fields:
                        if f in field:
                            vector_fields.append(field)
            print(
                f"The detected vector fields are {str(vector_fields)}, manually specify the `{field_type}_vector_fields` if those are incorrect."
            )
            if not vector_fields:
                raise (
                    f"No vector fields associated with the given {field_type} fields were found, run `ds.vectorize_{field_type}({field_type}_fields={str(fields)})` to extract vectors for your {field_type} fields."
                )
        return vector_fields

    def update_alias(self, field: str, alias: str):
        """Update the alias of a field using the SDK."""
        try:
            current_settings = self.datasets.get_settings(self.dataset_id)["results"]
        except:
            # If there are no settings
            self.datasets.post_settings(self.dataset_id, settings={})
            current_settings = self.datasets.get_settings(self.dataset_id)["results"]
        if "fieldAliases" not in current_settings:
            current_settings["fieldAliases"] = {}
        current_settings["fieldAliases"][field] = alias
        return self.datasets.post_settings(
            dataset_id=self.dataset_id, settings=current_settings
        )



import pytest

from typing import Any, Dict, List

from tests.globals.constants import NUMBER_OF_DOCUMENTS

from tests.globals.document import vector_document


@pytest.fixture(scope="session")
def vector_documents() -> List[Dict[str, Any]]:
    return [vector_document() for _ in range(NUMBER_OF_DOCUMENTS)]



from relevanceai.operations_new.cluster.batch.models.base import BatchClusterModelBase


class SklearnBatchClusterBase(BatchClusterModelBase):
    model: BatchClusterModelBase

    def partial_fit(self, X):
        return self.model.partial_fit(X)

    def predict(self, X):
        return self.model.predict(X)



from typing import List, Dict, Any, Optional, Union

from relevanceai.operations_new.transform_base import TransformBase
from relevanceai.operations_new.dr.models.base import DimReductionModelBase


class DimReductionTransform(TransformBase):

    model: DimReductionModelBase
    fields: List[str]
    alias: str

    def __init__(
        self,
        vector_fields: List[str],
        model: Union[str, DimReductionModelBase],
        n_components: int,
        alias: Optional[str] = None,
        model_kwargs: Optional[dict] = None,
        output_field: str = None,
        **kwargs,
    ):
        self.vector_fields = vector_fields

        # TODO: Add a get ailas method
        self.alias = alias  # type: ignore

        if model_kwargs is None:
            model_kwargs = {}

        self.model = self._get_model(
            model=model,
            n_components=n_components,
            alias=alias,
            **model_kwargs,
        )
        self.output_field = output_field
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def name(self):
        return "dimensionality_reduction"

    def _get_model(
        self,
        model: Union[str, DimReductionModelBase],
        n_components: int,
        alias: Union[str, None],
        **kwargs,
    ) -> DimReductionModelBase:
        """It returns a model.

        Parameters
        ----------
        model : Union[str, DimReductionModelBase]
            Union[str, DimReductionModelBase]
        n_components : int
            int
        alias : Union[str, None]
            Union[str, None]

        """
        if isinstance(model, str):
            model = self._get_model_from_string(
                model, n_components=n_components, alias=alias  # type: ignore
            )
            return model
        elif isinstance(model, DimReductionModelBase):
            return model
        else:
            raise ValueError(
                "dim reduction model provided should be either a string or inherit from DimReductionModelBase"
            )

    def _get_model_from_string(
        self, model: str, n_components: int, alias: str, **kwargs
    ) -> DimReductionModelBase:
        if model == "pca":
            from relevanceai.operations_new.dr.models.pca import PCAModel

            mapped_model = PCAModel(
                n_components=n_components,
                alias=alias,
                **kwargs,
            )

        elif model == "ivis":
            from relevanceai.operations_new.dr.models.ivis import IvisModel

            mapped_model = IvisModel(
                n_components=n_components,
                alias=alias,
                **kwargs,
            )

        elif model == "umap":
            from relevanceai.operations_new.dr.models.umap import UMAPModel

            mapped_model = UMAPModel(
                n_components=n_components,
                alias=alias,
                **kwargs,
            )

        elif model == "tsne":
            from relevanceai.operations_new.dr.models.tsne import TSNEModel

            mapped_model = TSNEModel(
                n_components=n_components,
                alias=alias,
                **kwargs,
            )

        else:
            raise ValueError(
                "relevanceai currently does not support this model as a string. the current supported models are [pca, tsne, umap, ivis]"
            )
        return mapped_model

    def transform(
        self,
        documents: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:

        concat_vectors: List[List[float]] = [[] for _ in range(len(documents))]

        for vector_field in self.vector_fields:
            vectors = self.get_field_across_documents(
                field=vector_field,
                docs=documents,
                missing_treatment="return_empty_string",
            )
            vector_length = len(
                self.get_field_across_documents(
                    field=vector_field,
                    docs=documents,
                    missing_treatment="skip",
                )[0]
            )

            for vector, concat_vector in zip(vectors, concat_vectors):
                if len(vector) == 0:
                    vector = [0] * vector_length
                concat_vector.extend(vector)

        reduced_vectors = self.model.fit_transform(concat_vectors)
        reduced_vector_name = self.model.vector_name(
            self.vector_fields, self.output_field
        )

        if reduced_vector_name in self.vector_fields:
            raise ValueError(
                "Alias is already being used, Please set a different alias"
            )

        self.set_field_across_documents(
            field=reduced_vector_name,
            values=reduced_vectors,
            docs=documents,
        )

        # removes unnecessary info for updated_where
        updated_documents = [
            {
                key: value
                for key, value in document.items()
                if key in ["_id", reduced_vector_name]
            }
            for document in documents
        ]

        return updated_documents



import time
from relevanceai.dataset import Dataset

from sklearn.cluster import KMeans


class TestOperation:
    def test_subcluster(self, test_dataset: Dataset):
        model = KMeans(n_clusters=4)

        vector_field = "sample_1_vector_"
        alias = "cluster_test_1"
        test_dataset.cluster(
            model=model,
            vector_fields=[vector_field],
            alias=alias,
            include_cluster_report=False,
        )
        time.sleep(1)
        assert f"_cluster_.{vector_field}.{alias}" in test_dataset.schema

        parent_field = f"_cluster_.{vector_field}.{alias}"
        # vector_field = "sample_2_vector_"
        alias = "subcluster_test_1"
        test_dataset.subcluster(
            model=model,
            alias=alias,
            vector_fields=[vector_field],
            parent_field=parent_field,
            filters=[
                {
                    "field": vector_field,
                    "filter_type": "exists",
                    "condition": "==",
                    "condition_value": "",
                }
            ],
            min_parent_cluster_size=4,
        )
        time.sleep(1)
        assert f"_cluster_.{vector_field}.{alias}" in test_dataset.schema



from typing import Any, Dict, Union

CLIP_MODELS: Dict[str, Dict[str, Any]] = {
    "clip": {
        "vector_length": 512,
        "url": "ViT-B/32",
    },
    "ViT-B/16": {
        "vector_length": 512,
        "url": "ViT-B/16",
    },
    "ViT-B/32": {
        "vector_length": 512,
        "url": "ViT-B/32",
    },
    "ViT-L/14": {
        "vector_length": 768,
        "url": "ViT-L/14",
    },
    "ViT-L/14@336px": {
        "vector_length": 768,
        "url": "ViT-L/14@336px",
    },
    "RN50": {
        "vector_length": 1024,
        "url": "RN50",
    },
    "RN101": {"vector_length": 512, "url": "RN101"},
    "RN50x4": {"vector_length": 640, "url": "RN50x4"},
    "RN50x16": {"vector_length": 512, "url": "RN50x16"},
    "RN50x64": {"vector_length": 1024, "url": "RN50x64"},
}

TFHUB_MODELS: Dict[str, Dict[str, Any]] = {
    "bit": {
        "vector_length": 2048,
        "url": "https://tfhub.dev/google/bit/s-r50x1/1",
    },
    "s-r50x1": {
        "vector_length": 2048,
        "url": "https://tfhub.dev/google/bit/s-r50x1/1",
    },
    "s-r50x3": {
        "vector_length": 6144,
        "url": "https://tfhub.dev/google/bit/s-r50x3/1",
    },
    "s-r101x1": {
        "vector_length": 2048,
        "url": "https://tfhub.dev/google/bit/s-r101x1/1",
    },
    "s-r101x3": {
        "vector_length": 6144,
        "url": "https://tfhub.dev/google/bit/s-r101x3/1",
    },
    "s-r152x4": {
        "vector_length": 8192,
        "url": "https://tfhub.dev/google/bit/s-r152x4/1",
    },
    "m-r50x1": {
        "vector_length": 2048,
        "url": "https://tfhub.dev/google/bit/m-r50x1/1",
    },
    "m-r50x3": {
        "vector_length": 6144,
        "url": "https://tfhub.dev/google/bit/m-r50x3/1",
    },
    "m-r101x1": {
        "vector_length": 2048,
        "url": "https://tfhub.dev/google/bit/m-r101x1/1",
    },
    "m-r101x3": {
        "vector_length": 6144,
        "url": "https://tfhub.dev/google/bit/m-r101x3/1",
    },
    "m-r152x4": {
        "vector_length": 8192,
        "url": "https://tfhub.dev/google/bit/m-r152x4/1",
    },
    "mobilenet_v1_100_224": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_100_224/feature_vector/4",
        "vector_length": 1024,
        "image_dimensions": 224,
    },
    "mobilenet_v1_100_192": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_100_192/feature_vector/4",
        "vector_length": 1024,
        "image_dimensions": 192,
    },
    "mobilenet_v1_100_160": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_100_160/feature_vector/4",
        "vector_length": 1024,
        "image_dimensions": 160,
    },
    "mobilenet_v1_100_128": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_100_128/feature_vector/4",
        "vector_length": 1024,
        "image_dimensions": 128,
    },
    "mobilenet_v1_075_224": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_075_224/feature_vector/4",
        "vector_length": 768,
        "image_dimensions": 224,
    },
    "mobilenet_v1_075_192": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_075_192/feature_vector/4",
        "vector_length": 768,
        "image_dimensions": 192,
    },
    "mobilenet_v1_075_160": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_075_160/feature_vector/4",
        "vector_length": 768,
        "image_dimensions": 160,
    },
    "mobilenet_v1_075_128": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_075_128/feature_vector/4",
        "vector_length": 768,
        "image_dimensions": 128,
    },
    "mobilenet_v1_050_224": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_050_224/feature_vector/4",
        "vector_length": 512,
        "image_dimensions": 224,
    },
    "mobilenet_v1_050_192": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_050_192/feature_vector/4",
        "vector_length": 512,
        "image_dimensions": 192,
    },
    "mobilenet_v1_050_160": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_050_160/feature_vector/4",
        "vector_length": 512,
        "image_dimensions": 160,
    },
    "mobilenet_v1_050_128": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_050_128/feature_vector/4",
        "vector_length": 512,
        "image_dimensions": 128,
    },
    "mobilenet_v1_025_224": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_025_224/feature_vector/4",
        "vector_length": 256,
        "image_dimensions": 224,
    },
    "mobilenet_v1_025_192": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_025_192/feature_vector/4",
        "vector_length": 256,
        "image_dimensions": 192,
    },
    "mobilenet_v1_025_160": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_025_160/feature_vector/4",
        "vector_length": 256,
        "image_dimensions": 160,
    },
    "mobilenet_v1_025_128": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_025_128/feature_vector/4",
        "vector_length": 256,
        "image_dimensions": 128,
    },
    "mobilenet": {
        "url": "https://tfhub.dev/google/imagenet/mobilenet_v1_025_128/feature_vector/4",
        "vector_length": 256,
        "image_dimensions": 128,
    },
    "inception_resnet_v2": {
        "url": "https://tfhub.dev/google/imagenet/inception_resnet_v2/feature_vector/4",
        "vector_length": 1536,
    },
    "inception_resnet_v2": {
        "url": "https://tfhub.dev/google/imagenet/inception_resnet_v2/feature_vector/4",
        "vector_length": 1536,
    },
    "inception_v1": {
        "vector_length": 1024,
        "url": "https://tfhub.dev/google/imagenet/inception_v1/feature_vector/4",
    },
}



import io
import uuid
import requests
import numpy as np
from relevanceai.apps.report_app.marks import ReportMarks


class ReportBlocks(ReportMarks):
    def _process_content(self, content):
        # Needs to be done better
        if isinstance(content, str):
            return [{"type": "text", "text": content}]
        elif isinstance(content, (float, int, np.generic)):
            return [{"type": "text", "text": str(content)}]
        elif isinstance(content, list):
            content_list = []
            for c in content:
                if isinstance(c, list) and len(c) == 1:
                    content_list.append(c[0])
                else:
                    processed_content = self._process_content(c)
                    if not isinstance(processed_content, list):
                        processed_content = [processed_content]
                    content_list += processed_content
            return content_list
        elif isinstance(content, dict):
            return [content]
        else:
            print(type(content))
            return content

    def markdown(self):
        pass

    def h1(self, content, add=True):
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "heading",
                    "attrs": {"level": 1},
                    "content": self._process_content(content),
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def h2(self, content, add=True):
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "heading",
                    "attrs": {"level": 2},
                    "content": self._process_content(content),
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def h3(self, content, add=True):
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "heading",
                    "attrs": {"level": 3},
                    "content": self._process_content(content),
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def quote(self, content, add=True):
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {"type": "blockquote", "content": self._process_content(content)}
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def code(self, content, language="python", add=True):
        block = {
            "type": "codeBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "attrs": {"language": language},
            "content": [
                {"type": "blockquote", "content": self._process_content(content)}
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def paragraph(self, content, add=True, raw=False):
        p_block = {"type": "paragraph", "content": self._process_content(content)}
        if raw:
            return p_block
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [p_block],
        }
        if add:
            self.contents.append(block)
        return block

    p = paragraph

    def space(self, height: int = 40, add=True):
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {"type": "spaceBlock", "attrs": {"width": "100%", "height": height}}
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def tooltip(self, content, tooltip_text, add=True):
        # has to be text block here.
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "tooltip",
                    "attrs": {"content": tooltip_text},
                    "content": self._process_content(content),
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def _list_item(self, content):
        return [{"type": "listItem", "content": [self.paragraph(content, raw=True)]}]

    def bullet_list(self, contents, add=True):
        if not isinstance(contents, list):
            raise TypeError("'contents' needs to be a List")
        list_contents = []
        for c in contents:
            list_contents += self._list_item(c)
        block = {
            "type": "appBlock",
            "content": [{"type": "bulletList", "content": list_contents}],
        }
        if add:
            self.contents.append(block)
        return block

    def ordered_list(self, contents, add=True):
        if not isinstance(contents, list):
            raise TypeError("'contents' needs to be a List")
        list_contents = []
        for c in contents:
            list_contents += self._list_item(c)
        block = {
            "type": "appBlock",
            "content": [{"type": "orderedList", "content": list_contents}],
        }
        if add:
            self.contents.append(block)
        return block

    def details(self, title_content, contents, collapsed: bool = True, add=True):
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "details",
                    "attrs": {"open": collapsed},
                    "content": [
                        {
                            "type": "detailsSummary",
                            "content": self._process_content(title_content),
                        },
                        {
                            "type": "detailsContent",
                            "content": self._process_content(contents),
                        },
                    ],
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def card(self, contents, width, color, add=True):
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "cardBlock",
                    "attrs": {"width": width, "colour": color},
                    "content": self._process_content(contents),
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def _column_content(self, content):
        return [{"type": "columnContent", "content": self._process_content(content)}]

    def columns(self, contents, num_columns: int = 2, add=True):
        if not isinstance(contents, list):
            raise TypeError("'contents' needs to be a List")
        list_contents = []
        for c in contents:
            list_contents += self._column_content(c)
        block = {
            "type": "appBlock",
            "content": [
                {
                    "type": "columnBlock",
                    "attrs": {"columns": num_columns},
                    "content": list_contents,
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    # def table(self, data, add=True):
    #     if data:
    #         table_headers = data.columns
    #         table_rows =
    #     block = {
    #         "type": "appBlock",
    #         "content": [{"type": "table", "content": table_rows}],
    #     }
    #     if add:
    #         self.contents.append(block)
    #     return block

    def image(
        self, content, title: str = "", width_percentage: int = 100, add: bool = True
    ):
        if isinstance(content, str):
            if "http" in content and "/":
                # online image
                content_bytes = io.BytesIO(requests.get(content).content).getvalue()
            else:
                # local filepath
                content_bytes = io.BytesIO(open(content, "rb").read()).getvalue()
        elif isinstance(content, bytes):
            content_bytes = content
        elif isinstance(content, io.BytesIO):
            content_bytes = content.getvalue()
        else:
            raise TypeError("'content' needs to be of type str, bytes or io.BytesIO.")
        filename = f"{title}.png" if title else f"{str(uuid.uuid4())}.png"
        image_url = self.dataset.insert_media_bytes(
            content_bytes, filename=filename, verbose=False
        )
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "imageDisplayBlock",
                    "attrs": {
                        "imageSrc": image_url,
                        "title": title,
                        "width": f"{width_percentage}%",
                        "height": "auto",
                    },
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def audio(
        self, content, title: str = "", width_percentage: int = 100, add: bool = True
    ):
        if isinstance(content, str):
            if "http" in content and "/":
                # online audio
                content_bytes = io.BytesIO(requests.get(content).content).getvalue()
            else:
                # local filepath
                content_bytes = io.BytesIO(open(content, "rb").read()).getvalue()
        elif isinstance(content, bytes):
            content_bytes = content
        elif isinstance(content, io.BytesIO):
            content_bytes = content.getvalue()
        else:
            raise TypeError("'content' needs to be of type str, bytes or io.BytesIO.")
        filename = f"{title}.wav" if title else f"{str(uuid.uuid4())}.wav"
        audio_url = self.dataset.insert_media_bytes(
            content_bytes, filename=filename, verbose=False
        )
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "audioDisplayBlock",
                    "attrs": {
                        "audioSrc": audio_url,
                        "title": title,
                        "width": f"{width_percentage}%",
                        "height": "auto",
                    },
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block

    def video(
        self, content, title: str = "", width_percentage: int = 100, add: bool = True
    ):
        if isinstance(content, str):
            if "http" in content and "/":
                # online video
                content_bytes = io.BytesIO(requests.get(content).content).getvalue()
            else:
                # local filepath
                content_bytes = io.BytesIO(open(content, "rb").read()).getvalue()
        elif isinstance(content, bytes):
            content_bytes = content
        elif isinstance(content, io.BytesIO):
            content_bytes = content.getvalue()
        else:
            raise TypeError("'content' needs to be of type str, bytes or io.BytesIO.")
        filename = f"{title}.mp4" if title else f"{str(uuid.uuid4())}.mp4"
        video_url = self.dataset.insert_media_bytes(
            content_bytes, filename=filename, verbose=False
        )
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content": [
                {
                    "type": "videoDisplayBlock",
                    "attrs": {
                        "videoSrc": video_url,
                        "title": title,
                        "width": f"{width_percentage}%",
                        "height": "auto",
                    },
                }
            ],
        }
        if add:
            self.contents.append(block)
        return block



from relevanceai.operations.dr.models.ivis import Ivis
from relevanceai.operations.dr.models.pca import PCA
from relevanceai.operations.dr.models.umap import UMAP
from relevanceai.operations.dr.models.tsne import TSNE



from typing import List, Union

import numpy as np

from relevanceai.operations_new.dr.models.base import DimReductionModelBase

from sklearn.decomposition import PCA


class PCAModel(DimReductionModelBase):
    def __init__(
        self,
        n_components: int,
        alias: Union[str, None],
        **kwargs,
    ):
        self.model = PCA(n_components=n_components, **kwargs)
        self.model_name = "pca"
        self.alias = alias

    def fit(
        self,
        vectors: Union[List[List[float]], np.ndarray],
    ) -> None:
        """It fits the model to the vectors.

        Parameters
        ----------
        vectors : Union[List[List[float]], np.ndarray]
            Union[List[List[float]], np.ndarray]

        """

        if isinstance(vectors, list):
            vectors = np.array(vectors)

        self.model.fit(vectors)

    def fit_transform(
        self,
        vectors: List[List[float]],
    ) -> List[List[float]]:
        """It takes a list of vectors, fits the model to the vectors, and then transforms the vectors

        Parameters
        ----------
        vectors : List[List[float]]
            List[List[float]]

        Returns
        -------
            A list of lists of floats.

        """

        vectors = np.array(vectors)
        reduced_vectors = self.model.fit_transform(vectors)
        return reduced_vectors.tolist()



from relevanceai.operations_new.processing.text.clean.helpers import (
    BaseTextProcessing,
    MLStripper,
)

from relevanceai.operations_new.transform_base import TransformBase


class CleanTextTransform(TransformBase):
    def __init__(
        self,
        text_fields: list,
        output_fields: list,
        remove_html_tags: bool = True,
        lower=False,
        remove_punctuation=True,
        remove_digits=True,
        remove_stopwords: list = None,
        lemmatize: bool = False,
        replace_words: dict = None,
        **kwargs
    ):
        if len(text_fields) != len(output_fields):
            raise ValueError("Text fields and output fields are not equal!")
        self.text_fields = text_fields
        self.output_fields = output_fields
        self.remove_html_tags = remove_html_tags
        self.lower = lower
        self.remove_punctuation = remove_punctuation
        self.remove_digits = remove_digits
        self.remove_stopwords = remove_stopwords
        self.lemmatize = lemmatize
        self.replace_words = replace_words
        # Set all the other kwargs!
        for k, v in kwargs.items():
            setattr(self, k, v)

    def clean_text(self, text):
        """
        Clean the text of the individuals
        """
        # TODO: Make this faster by moving the operations into the folder
        try:
            if self.lower:
                text = BaseTextProcessing.lower_text(text)
            if self.remove_punctuation:
                text = BaseTextProcessing.remove_punctuation(text)
            if self.remove_html_tags:
                text = BaseTextProcessing.remove_html_tags(text)
            if self.remove_digits:
                text = BaseTextProcessing.remove_digits(text)
            if self.remove_stopwords:
                text = BaseTextProcessing.remove_stopwords(
                    text, additional_stp_wrds=self.remove_stopwords
                )
            if self.lemmatize:
                text = BaseTextProcessing.lemmatize(text)
            if self.replace_words:
                text = BaseTextProcessing.replace_words(text, self.replace_words)
        except Exception as e:
            import traceback

            traceback.print_exc()
        return text

    def clean_text_document(self, text_field, document, output_field):
        """
        Split a text field and store it in other values
        """
        t = self.get_field(
            text_field, document, missing_treatment="return_empty_string"
        )
        clean_text = self.clean_text(t)
        # Format the split text into documents
        new_doc = {"_id": document["_id"]}
        self.set_field(output_field, new_doc, clean_text)
        return new_doc

    def clean_text_documents(self, text_field, documents, output_field):
        return [
            self.clean_text_document(
                text_field=text_field, document=document, output_field=output_field
            )
            for document in documents
        ]

    def transform(self, documents):
        for i, t in enumerate(self.text_fields):
            new_documents = self.clean_text_documents(
                t, documents, self.output_fields[i]
            )
        return new_documents

    def name(self):
        return 'clean_text'



"""
Faiss Base
"""
from typing import Any, List, Dict, Optional, Tuple

import numpy as np

from relevanceai.operations_new.cluster.models.base import ClusterModelBase

from faiss import Kmeans


class FaissModel(ClusterModelBase):
    """Faiss model base"""

    model: Kmeans

    def __init__(
        self,
        model: Any,
        model_kwargs: Optional[Dict] = None,
    ):

        if isinstance(model, str):
            model = ClusterModelBase.import_from_string(f"faiss.Kmeans")
            if model_kwargs is None:
                model_kwargs = {}

            self.model = model(**model_kwargs)
        else:
            self.model = model

    @staticmethod
    def check_vectors(input: Any) -> np.ndarray:
        """It takes a list of vectors, converts them to a numpy array, and then converts them to a float32

        Parameters
        ----------
        input : Any
            Any

        Returns
        -------
            the input as a numpy array.

        """
        if not isinstance(input, np.ndarray):
            try:
                input = np.array(input)
            except Exception:
                raise ValueError("Could not convert vectors to np.ndarray")

        try:
            output = input.astype(np.float32)
        except Exception:
            raise ValueError("Could not convert vectors to np.float32")

        return output

    def fit_predict(self, vectors: Any) -> List[int]:
        """> The function takes in a list of vectors and returns a list of labels

        Parameters
        ----------
        vectors : Any
            Any - the vectors to be clustered

        Returns
        -------
            The labels of the vectors.

        """
        vectors = FaissModel.check_vectors(vectors)

        self.fit(vectors)
        res: Tuple[np.ndarray, np.ndarray] = self.model.assign(vectors)

        labels: np.ndarray = res[1]

        return labels.tolist()

    def fit(self, vectors: Any) -> np.float64:
        """It trains the model.

        Parameters
        ----------
        vectors : Any
            Any

        Returns
        -------
            The model is being trained on the vectors.

        """
        vectors = FaissModel.check_vectors(vectors)

        return self.model.train(vectors)

    def predict(self, vectors) -> List[int]:
        """This function takes in a list of vectors and returns a list of labels

        Parameters
        ----------
        vectors
            The vectors to be predicted.

        Returns
        -------
            The labels of the vectors.

        """
        vectors = FaissModel.check_vectors(vectors)

        res: Tuple[np.ndarray, np.ndarray] = self.model.assign(vectors)

        labels: np.ndarray = res[1]

        return labels.tolist()

    @property
    def cluster_centers_(self):
        """It returns the centroids of the clusters

        Returns
        -------
            The cluster centers

        """
        return self.model.centroids

    @property
    def alias(self):
        """The function takes a string and returns a string

        Returns
        -------
            The alias of the model.

        """
        return f"fiass{self.name}-{str(self.model.k)}"



from relevanceai.apps.report_app.advanced_blocks.connected_charts import (
    ConnectedChartsBlocks,
)
from relevanceai.apps.report_app.advanced_blocks.pyplot import PyplotReportBlock
from relevanceai.apps.report_app.advanced_blocks.plotly import PlotlyReportBlock
from relevanceai.apps.report_app.advanced_blocks.altair import AltairReportBlock
from relevanceai.apps.report_app.advanced_blocks.markdown import MarkdownBlock


class ReportAdvancedBlocks(
    PyplotReportBlock, PlotlyReportBlock, AltairReportBlock, MarkdownBlock
):
    def plot_by_method(
        self, plot, plot_method, title="", height=None, width=None, add=True
    ):
        if plot_method == "plotly":
            self.plotly(plot, title=title, height=height, width=width, add=add)
        elif plot_method == "pyplot":
            self.pyplot(plot, title=title, height=height, width=width, add=add)

    # def plot_dendrogram()



from relevanceai.operations_new.emotion.transform import EmotionTransform
from relevanceai.operations_new.ops_base import OperationAPIBase


class EmotionOps(OperationAPIBase, EmotionTransform):
    """
    Add emotions
    """

    def __init__(
        self,
        text_fields: list,
        model_name="joeddav/distilbert-base-uncased-go-emotions-student",
        output_fields: list = None,
        min_score: float = 0.3,
        **kwargs,
    ):
        self.model_name = model_name
        self.text_fields = text_fields
        if output_fields is not None:
            self.output_fields = output_fields
        else:
            self.output_fields = [self._generate_output_field(t) for t in text_fields]
        self.min_score = min_score
        super().__init__(**kwargs)

    @property
    def name(self):
        return "emotion"



from typing import List, Any

from relevanceai.operations_new.vectorize.models.base import VectorizeModelBase
from relevanceai.utils.decorators.vectors import catch_errors


class TFHubText2Vec(VectorizeModelBase):
    def __init__(self, url, vector_length):
        import tensorflow_hub as hub

        self.model: Any = hub.load(url)
        self.vector_length: int = vector_length
        self.model_name = self._get_model_name(url)

    @catch_errors
    def encode(self, text: str) -> List[float]:
        """The function takes a string as input and returns a list of floats

        Parameters
        ----------
        text : str
            The text to be encoded.

        Returns
        -------
            A list of floats.

        """
        return self.model.encode([text])[0].tolist()

    @catch_errors
    def bulk_encode(self, texts: List[str]) -> List[List[float]]:
        """The function takes a list of strings and returns a list of lists of floats

        Parameters
        ----------
        texts : List[str]
            List[str]

        Returns
        -------
            A list of lists of floats.

        """
        return self.model.encode(texts).tolist()



import warnings
from relevanceai.apps.report_app.base import ReportBase


class ReportMarks(ReportBase):
    def _process_marks(self, content, mark, nested=False):
        #Needs to be done better
        if isinstance(content, list):
            new_content = []
            for c in content:
                if isinstance(c, list) and len(c) == 1:
                    new_content.append(self._process_marks(c[0], mark, nested=True))
                else:
                    new_content.append(self._process_marks(c, mark, nested=True))
            return new_content
        elif isinstance(content, dict):
            if "marks" in content:
                content['marks'] += mark
            return content
        elif nested:
            return {"type": "text", "text": content, "marks": mark}
        else:
            return [{"type": "text", "text": content, "marks": mark}]

    def bold(self, content):
        return self._process_marks(content, [{"type": "bold"}])

    def italic(self, content):
        return self._process_marks(content, [{"type": "italic"}])

    def strike(self, content):
        return self._process_marks(content, [{"type": "strike"}])

    def underline(self, content):
        return self._process_marks(content, [{"type": "underline"}])

    def inline_code(self, content):
        return self._process_marks(content, [{"type": "code"}])

    def highlight(self, content, start, end, color):
        return self._process_marks(content, [
            {
                "type": "highlight",
                "attrs": {"color": color},
                "from": start,
                "to": end,
            }
        ])

    def color(self, content, color, background_color):
        return self._process_marks(content, [
            {
                "type": "textStyle",
                "attrs": {"color": color, "backgroundColor": background_color},
            }
        ])

    def link(self, content, href):
        return self._process_marks(content, [
            {"type": "link", "attrs": {"href": href, "target": "_blank"}}
        ])



"""
All operations related to concept_graph
"""
from relevanceai.operations_new.sentiment.transform import SentimentTransform
from relevanceai.operations_new.ops_base import OperationAPIBase


class SentimentOps(SentimentTransform, OperationAPIBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



"""
    Run operations to get the keyphrases of this document
"""

from relevanceai.operations_new.processing.text.keywords.transform import (
    KeyWordTransform,
)
from relevanceai.operations_new.ops_base import OperationAPIBase


class KeyWordOps(OperationAPIBase, KeyWordTransform):
    def __init__(
        self,
        fields: list,
        model_name: str = "all-mpnet-base-v2",
        lower_bound: int = 0,
        upper_bound: int = 3,
        output_fields: list = None,
        stop_words: list = None,
        max_keywords: int = 1,
        use_maxsum: bool = False,
        nr_candidates: int = 20,
        **kwargs
    ):
        self.fields = fields
        self.model_name = model_name
        self.output_fields = output_fields
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.stop_words = stop_words
        self.max_keywords = max_keywords
        self.use_maxsum = use_maxsum
        self.nr_candidates = nr_candidates
        super().__init__(**kwargs)



"""
All read operations for Dataset
"""
import collections
import re
import math
import warnings
import os
import pandas as pd
import uuid
from relevanceai.utils.decorators.thread import fire_and_forget

from tqdm.auto import tqdm
from typing import Any, Optional, Tuple, Union, Dict, List, Mapping
from relevanceai.client.helpers import Credentials
from relevanceai.dataset.series import Series


from relevanceai.dataset.read.metadata import Metadata
from relevanceai.dataset.read.cluster import ClusterRead
from relevanceai.dataset.helpers import _build_filters

from relevanceai.utils.cache import lru_cache
from relevanceai.utils.decorators.analytics import track

from relevanceai.constants.constants import MAX_CACHESIZE
from relevanceai.constants.warning import Warning


def update_nested_dictionary(d: dict, u: Union[dict, Mapping]):
    """ "If the value is a dictionary, recursively call update_nested_dictionary on it, otherwise just set
    the value."

    The function takes two dictionaries as arguments, d and u. It iterates over the key-value pairs in
    u, and for each pair, it checks if the value is a dictionary. If it is, it calls
    update_nested_dictionary on the value and the corresponding value in d. If the value is not a
    dictionary, it just sets the value in d to the value in u

    Parameters
    ----------
    d : dict
        The dictionary to update
    u : dict
        the dictionary

    Returns
    -------
        A dictionary with the updated values.

    """
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            d[k] = update_nested_dictionary(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class Read(ClusterRead):
    """

    Dataset Read
    -------------------

    A Pandas Like datatset API for interacting with the RelevanceAI python package
    """

    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
        fields: Optional[list] = None,
        image_fields: Optional[List[str]] = None,
        audio_fields: Optional[List[str]] = None,
        text_fields: Optional[List[str]] = None,
        highlight_fields: Optional[Dict[str, list]] = None,
        **kwargs,
    ):

        from relevanceai.operations.cluster.centroids import Centroids

        self.credentials = credentials
        self.fields = [] if fields is None else fields
        self.dataset_id = dataset_id

        self.centroids = Centroids(
            credentials=credentials,
            dataset_id=self.dataset_id,
        )
        self.image_fields = [] if image_fields is None else image_fields
        self.audio_fields = [] if audio_fields is None else audio_fields
        self.text_fields = [] if text_fields is None else text_fields
        self.highlight_fields = {} if highlight_fields is None else highlight_fields
        super().__init__(
            credentials=credentials,
            dataset_id=dataset_id,
            **kwargs,
        )

    @property  # type: ignore
    @track
    def shape(self):
        """
        Returns the shape (N x C) of a dataset
        N = number of samples in the Dataset
        C = number of columns in the Dataset

        Returns
        -------
        Tuple
            (N, C)

        Example
        ---------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            length, width = df.shape
        """
        schema = self.schema
        n_documents = self.get_number_of_documents(dataset_id=self.dataset_id)
        return (n_documents, len(schema))

    def _get_possible_dtypes(self, schema):
        possible_dtypes = []
        for v in schema.values():
            if isinstance(v, str):
                possible_dtypes.append(v)
            elif isinstance(v, dict):
                if list(v)[0] == "vector":
                    possible_dtypes.append("vector_")
        return possible_dtypes

    def _get_dtype_count(self, schema: dict):
        possible_dtypes = self._get_possible_dtypes(schema)
        dtypes = {
            dtype: list(schema.values()).count(dtype) for dtype in possible_dtypes
        }
        return dtypes

    def _get_schema(self):
        # stores schema in memory to save users API usage/reloading
        if hasattr(self, "_schema"):
            return self._schema
        self._schema = self.datasets.schema(self.dataset_id)
        return self._schema

    @track
    def info(self, dtype_count: bool = False) -> pd.DataFrame:
        """
        Return a dictionary that contains information about the Dataset
        including the index dtype and columns and non-null values.

        Parameters
        -----------
        dtype_count: bool
            If dtype_count is True, prints a value_counts of the data type


        Returns
        ---------
        pd.DataFrame
            a pandas dataframe of information

        Example
        ---------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.info()
        """
        health: dict = self.datasets.monitor.health(self.dataset_id)
        schema: dict = self._get_schema()
        info_json = [
            {
                "Column": column,
                "Dtype": schema[column],
            }
            if column not in health
            else {
                "Column": column,
                "Non-Null Count": health[column]["missing"],
                "Dtype": schema[column],
            }
            for column in schema
        ]

        info_df = pd.DataFrame(info_json)
        if dtype_count:
            dtypes_info = self._get_dtype_count(schema)
            print(dtypes_info)
        return info_df

    @track
    def head(
        self, n: int = 5, raw_json: bool = False, select_fields: list = None, **kw
    ) -> Union[dict, pd.DataFrame]:
        """
        Return the first `n` rows.
        returns the first `n` rows of your dataset.
        It is useful for quickly testing if your object
        has the right type of data in it.

        Parameters
        ----------
        n : int, default 5
            Number of rows to select.
        raw_json: bool
            If True, returns raw JSON and not Pandas Dataframe
        kw:
            Additional arguments to feed into show_json

        Returns
        -------
        Pandas DataFrame or Dict, depending on args
            The first 'n' rows of the caller object.

        Example
        ---------
        .. code-block::

            from relevanceai import Client

            client = Client()

            df = client.Dataset("sample_dataset_id", image_fields=["image_url])

            df.head()
        """
        print(
            f"https://cloud.relevance.ai/dataset/{self.dataset_id}/dashboard/data?page=1"
        )
        head_documents = self.get_documents(
            number_of_documents=n, include_after_id=False, select_fields=select_fields
        )
        if raw_json:
            return head_documents
        else:
            try:
                return self._show_json(head_documents, **kw)
            except Exception as e:
                warnings.warn(Warning.MISSING_RELEVANCE_NOTEBOOK + str(e))
                return pd.json_normalize(head_documents).head(n=n)

    def _show_json(self, documents, **kw):
        from jsonshower import show_json

        if not self.text_fields:
            text_fields = pd.json_normalize(documents).columns.tolist()
        else:
            text_fields = self.text_fields
        return show_json(
            documents,
            image_fields=self.image_fields,
            audio_fields=self.audio_fields,
            highlight_fields=self.highlight_fields,
            text_fields=text_fields,
            **kw,
        )

    def _repr_html_(self):
        documents = self.get_documents(include_after_id=False)
        documents = [
            {
                "_id": document["_id"],
                "insert_date_": document["insert_date_"],
                **document,
            }
            for document in documents
        ]
        try:
            return self._show_json(documents, return_html=True)
        except Exception as e:
            warnings.warn(Warning.MISSING_RELEVANCE_NOTEBOOK + str(e))
            return pd.json_normalize(documents).set_index("_id")._repr_html_()

    @track
    def sample(
        self,
        n: int = 1,
        frac: float = None,
        filters: Optional[list] = None,
        random_state: int = 0,
        select_fields: Optional[list] = None,
        include_vector: bool = True,
        output_format: str = "json",
    ):
        """
        Return a random sample of items from a dataset.

        Parameters
        ----------
        n : int
            Number of items to return. Cannot be used with frac.
        frac: float
            Fraction of items to return. Cannot be used with n.
        filters: list
            Query for filtering the search results
        random_state: int
            Random Seed for retrieving random documents.
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id", image_fields=["image_url])
            df.sample()
        """
        filters = [] if filters is None else filters
        select_fields = [] if select_fields is None else select_fields

        if not select_fields and self.fields:
            select_fields = self.fields

        if frac and n:
            raise ValueError("Only one of n or frac can be provided")

        if frac:
            if frac > 1 or frac < 0:
                raise ValueError("Fraction must be between 0 and 1")
            n = math.ceil(
                self.get_number_of_documents(self.dataset_id, filters=filters) * frac
            )

        documents = self.datasets.documents.get_where(
            dataset_id=self.dataset_id,
            filters=filters,
            page_size=n,
            random_state=random_state,
            is_random=True,
            select_fields=select_fields,
            include_vector=include_vector,
        )["documents"]
        if output_format == "json":
            return documents
        elif output_format == "pandas":
            return pd.DataFrame.from_dict(documents, orient="records")

    @lru_cache(maxsize=MAX_CACHESIZE)
    @track
    def get_all_documents(
        self,
        chunksize: int = 1000,
        filters: Optional[List] = None,
        sort: Optional[List] = None,
        select_fields: Optional[List] = None,
        include_vector: bool = True,
        show_progress_bar: bool = True,
    ):
        """
        Retrieve all documents with filters. Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched. For more details see documents.get_where.

        Parameters
        ------------
        chunksize: list
            Number of documents to retrieve per retrieval
        include_vector: bool
            Include vectors in the search results
        sort: list
            Fields to sort by. For each field, sort by descending or ascending. If you are using descending by datetime, it will get the most recent ones.
        filters: list
            Query for filtering the search results
        select_fields : list
            Fields to include in the search results, empty array/list means all fields.

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            documents = df.get_all_documents()

        """
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort
        select_fields = [] if select_fields is None else select_fields

        return self._get_all_documents(
            dataset_id=self.dataset_id,
            chunksize=chunksize,
            filters=filters,
            sort=sort,
            select_fields=select_fields,
            include_vector=include_vector,
            show_progress_bar=show_progress_bar,
        )

    @track
    def get_documents_by_ids(
        self, document_ids: Union[List, str], include_vector: bool = True
    ):
        """
        Retrieve a document by its ID ("_id" field). This will retrieve the document faster than a filter applied on the "_id" field.

        Parameters
        ----------
        document_ids: Union[list, str]
            ID of a document in a dataset.
        include_vector: bool
            Include vectors in the search results

        Example
        --------

        .. code-block::

            from relevanceai import Client, Dataset
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.get_documents_by_ids(["sample_id"], include_vector=False)
        """
        if isinstance(document_ids, str):
            return self.datasets.documents.get(
                self.dataset_id, id=document_ids, include_vector=include_vector
            )
        elif isinstance(document_ids, list):
            return self.datasets.documents.bulk_get(
                self.dataset_id, ids=document_ids, include_vector=include_vector
            )
        raise TypeError("Document IDs needs to be a string or a list")

    @track
    def get(self, document_ids: Union[List, str], include_vector: bool = True):
        """
        Retrieve a document by its ID ("_id" field). This will retrieve the document faster than a filter applied on the "_id" field.
        This has the same functionality as get_document_by_ids.

        Parameters
        ----------
        document_ids: Union[list, str]
            ID of a document in a dataset.
        include_vector: bool
            Include vectors in the search results

        Example
        --------

        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.get(["sample_id"], include_vector=False)
        """
        if isinstance(document_ids, str):
            return self.datasets.documents.get(
                self.dataset_id, id=document_ids, include_vector=include_vector
            )
        elif isinstance(document_ids, list):
            return self.datasets.documents.bulk_get(
                self.dataset_id, ids=document_ids, include_vector=include_vector
            )
        raise TypeError("Document IDs needs to be a string or a list")

    @property  # type: ignore
    @track
    def schema(self) -> Dict:
        """
        Returns the schema of a dataset. Refer to datasets.create for different field types available in a Relevance schema.

        Example
        -----------------

        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.schema
        """
        return self.datasets.schema(self.dataset_id)

    @property  # type: ignore
    @track
    def columns(self) -> List[str]:
        """
        Returns a list of columns

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.columns

        """
        return list(self.schema)

    @track
    def filter(
        self,
        index: Union[str, None] = None,
        items: Union[List, None] = None,
        like: Union[str, None] = None,
        regex: Union[str, None] = None,
        axis: Union[int, str] = 0,
    ):
        """
        Returns a subset of the dataset, filtered by the parameters given

        Parameters
        ----------
        items : str, default None
            the column on which to filter, if None then defaults to the _id column
        items : list-like
            Keep labels from axis which are in items.
        like : str
            Keep labels from axis for which "like in label == True".
        regex : str (regular expression)
            Keep labels from axis for which re.search(regex, label) == True.
        axis : {0 or `index`, 1 or `columns`},
            The axis on which to perform the search

        Returns
        ---------
        list of documents

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("ecommerce-example-encoded")
            filtered = df.filter(items=["product_title", "query", "product_price"])
            filtered = df.filter(index="query", like="routers")
            filtered = df.filter(index="product_title", regex=".*Hard.*Drive.*")

        """
        fields = []
        filters = []

        schema = list(self.schema)

        if index:
            axis = 0
        else:
            axis = 1
            index = "_id"

        rows = axis in [0, "index"]
        columns = axis in [1, "columns"]

        if items is not None:
            if columns:
                fields += items

            elif rows:
                filters += _build_filters(items, filter_type="exact_match", index=index)

        elif like:
            if columns:
                fields += [column for column in schema if like in column]

            elif rows:
                filters += _build_filters(like, filter_type="contains", index=index)

        elif regex:
            if columns:
                query = re.compile(regex)
                re_fields = list(filter(query.match, schema))
                fields += re_fields

            elif rows:
                filters += _build_filters(regex, filter_type="regexp", index=index)

        else:
            raise TypeError("Must pass either `items`, `like` or `regex`")

        filters = [{"filter_type": "or", "condition_value": filters}]

        return self.get_all_documents(select_fields=fields, filters=filters)

    @track
    def get_documents(
        self,
        number_of_documents: int = 20,
        filters: Optional[list] = None,
        cursor: str = None,
        chunksize: int = 1000,
        sort: Optional[list] = None,
        select_fields: Optional[list] = None,
        include_vector: bool = True,
        include_cursor: bool = False,
        after_id: Optional[list] = None,
        include_after_id: bool = True,
    ):
        """
        Retrieve documents with filters. Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched. \n
        If you are looking to combine your filters with multiple ORs, simply add the following inside the query {"strict":"must_or"}.
        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        number_of_documents: int
            Number of documents to retrieve
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        cursor: string
            Cursor to paginate the document retrieval
        chunksize: int
            Number of documents to retrieve per iteration
        include_vector: bool
            Include vectors in the search results
        sort: list
            Fields to sort by. For each field, sort by descending or ascending. If you are using descending by datetime, it will get the most recent ones.
        filters: list
            Query for filtering the search results
        """
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort
        select_fields = [] if select_fields is None else select_fields

        return self._get_documents(
            dataset_id=self.dataset_id,
            number_of_documents=number_of_documents,
            filters=filters,
            cursor=cursor,
            chunksize=chunksize,
            sort=sort,
            select_fields=select_fields,
            include_vector=include_vector,
            include_cursor=include_cursor,
            after_id=after_id,
            include_after_id=include_after_id,
        )

    @track
    def get_metadata(self):
        """
        Store Metadata
        """
        try:
            return self.datasets.metadata(self.dataset_id)["results"]
        except:
            return {}

    @property
    def metadata(self):
        """Get the metadata"""
        _metadata = self.get_metadata()
        self._metadata = Metadata(_metadata, self.credentials, self.dataset_id)
        return self._metadata

    @track
    def insert_metadata(self, metadata: dict):
        """Insert metadata"""
        results = self.datasets.post_metadata(self.dataset_id, metadata)
        if results == {}:
            print("✅ You have successfully upserted metadata.")
        else:
            return results

    @track
    def upsert_metadata(self, metadata: dict):
        """Upsert metadata."""
        original_metadata: dict = self.get_metadata()
        update_nested_dictionary(original_metadata, metadata)
        results = self.datasets.post_metadata(self.dataset_id, metadata)
        if results == {}:
            print("✅ You have successfully inserted metadata.")
        else:
            return results

    def _update_workflow_progress(self, metadata: dict, workflow_id: str = None):
        """
        Updates the workflow progress
        """
        # RUn it asynchronously
        if workflow_id is None:
            workflow_id = os.getenv("WORKFLOW_ID", None)

        @fire_and_forget
        def update():
            return self.workflows.metadata(
                workflow_id=workflow_id, metadata=self.json_encoder(metadata)
            )

        if workflow_id is not None and workflow_id != "":
            update()

    @track
    def chunk_dataset(
        self,
        select_fields: List = None,
        chunksize: int = 100,
        filters: list = None,
        after_id: list = None,
    ):
        """

        Function for chunking a dataset

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            for c in ds.chunk_dataset(
                select_fields=["sample_label"],
                chunksize=100
            ):
                # Returns a dictionary with 'cursor' and 'documents' keys
                docs = c['documents']
                cursor = c['cursor']
                for d in docs:
                    d.update({"value": 3})
                ds.upsert_documents(docs)

        """
        docs = self.get_documents(
            number_of_documents=chunksize,
            filters=filters,
            select_fields=select_fields,
            include_after_id=True,
            after_id=after_id,
        )
        number_of_documents = self.get_number_of_documents(
            self.dataset_id, filters=filters
        )
        with tqdm(range(math.ceil(number_of_documents / chunksize))) as pbar:
            # update after we get the first batch
            pbar.update(1)
            self._update_workflow_progress(metadata=pbar.format_dict)
            while len(docs["documents"]) > 0:
                # Update metadata if possible
                # {'n': 0, 'total': 3, 'elapsed': 1.3828277587890625e-05,
                # 'ncols': 143, 'nrows': 23, 'prefix': '',
                # 'ascii': False, 'unit': 'it', 'unit_scale': False,
                # 'rate': None, 'bar_format': None, 'postfix': None,
                # 'unit_divisor': 1000, 'initial': 0, 'colour': None}
                yield docs["documents"]
                docs = self.get_documents(
                    number_of_documents=chunksize,
                    include_cursor=False,
                    after_id=docs["after_id"],
                    filters=filters,
                    select_fields=select_fields,
                )
                # Final update at the end
                pbar.update(1)
                self._update_workflow_progress(metadata=pbar.format_dict)
        return

    @track
    def list_vector_fields(self):
        """
        Returns list of valid vector fields in dataset
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("_mock_dataset_")
            ds.list_vector_fields()

        """
        schema = self.schema
        return [
            k for k in schema.keys() if k.endswith("_vector_") and "_cluster_" not in k
        ]

    @track
    def list_cluster_aliases(self):
        raise NotImplementedError()

    def isnull(
        self,
        show_progress_bar=False,
    ):
        schema = self.schema
        filters = [
            {
                "filter_type": "or",
                "condition_value": [
                    Series(
                        credentials=self.credentials,
                        dataset_id=self.dataset_id,
                        field=field,
                    ).not_exists()
                    for field in schema
                ],
            }
        ]
        null_count = {field: 0 for field in schema}

        null_docs = self.get_all_documents(
            filters=filters,
            show_progress_bar=show_progress_bar,
        )

        for doc in null_docs:
            for field in doc:
                if doc[field] is None:
                    null_count[field] += 1

        return null_count

    def facets(
        self,
        fields: list,
        date_interval: str = "monthly",
        page_size: int = 5,
        page: int = 1,
        asc: bool = False,
    ):
        return self.datasets.facets(
            dataset_id=self.dataset_id,
            fields=fields,
            date_interval=date_interval,
            page_size=page_size,
            page=page,
            asc=asc,
        )

    def get_settings(self):
        """
        Get the settings in dataset
        """
        return self.datasets.get_settings(dataset_id=self.dataset_id)

    def list_cluster_fields(self):
        return [x for x in self.schema if "_cluster_" in x and x.count(".") >= 2]

    def get_after_ids_for_workflows(
        self, num_of_workers: int = 3
    ) -> List[Tuple[int, List[Optional[str]]]]:
        """
        Get multiple after IDs to run workflows in parallel

        Params
        -------
        num_of_workers: int = 3
            The number of workers that we need to separate out the After IDs
        """
        documents: List[Dict[str, Any]] = [{"placeholder": "placeholder"}]
        docs = {"after_id": None}
        after_ids: List[Tuple[int, List[Optional[str]]]] = []
        after_id: List[Optional[str]] = []
        pull_limit = math.ceil(self.shape[0] / num_of_workers)

        while documents:
            docs = self.datasets.documents.get_where(
                dataset_id=self.dataset_id,
                page_size=pull_limit,
                select_fields=["_id"],
                after_id=after_id,
            )
            documents = docs["documents"]  # type: ignore
            pull_limit = len(documents)
            after_ids += [(pull_limit, after_id)]
            after_id = docs["after_id"]  # type: ignore

        return after_ids

    def update_field_children(
        self,
        field: str,
        field_children: list,
        category: str,
        metadata: dict = None,
    ):
        """
        Update the field children.
        """
        # Update field children
        if metadata is None:
            metadata = {}

        if "category" not in metadata:
            metadata["category"] = category
        return self.datasets.field_children.update(
            dataset_id=self.dataset_id,
            fieldchildren_id=uuid.uuid4().__str__(),
            field=field,
            field_children=field_children,
            metadata=metadata,
        )

    def list_field_children(
        self,
    ):
        """
        List field children
        """
        return self.datasets.field_children.list(dataset_id=self.dataset_id)

    def delete_field_children(
        self,
        fieldchildren_id: str,
    ):
        """
        Delete field children
        """
        return self.datasets.field_children.delete(
            dataset_id=self.dataset_id, fieldchildren_id=fieldchildren_id
        )



import numpy as np
from relevanceai._api.endpoints.datasets import cluster
from relevanceai.apps import ReportApp
from relevanceai.recipes.model_observability.cluster.evaluator import ClusterEvaluator
from typing import Union, List, Dict, Any, Optional


class ClusterReport(ReportApp):
    def __init__(self, name: str, dataset, deployable_id: str = None, **kwargs):
        self.started_cluster_evaluator = False
        super().__init__(
            dataset=dataset,
            name=name,
            deployable_id=deployable_id,
            **kwargs,
        )

    def start_cluster_evaluator(
        self,
        X: Union[list, np.ndarray],
        cluster_labels: Union[List[Union[str, float]], np.ndarray],
        centroids: Union[list, np.ndarray, str] = None,
        cluster_names: Union[list, dict] = None,
        feature_names: Union[list, dict] = None,
        model=None,
        outlier_label: Union[str, int] = -1,
        metric: str = "euclidean",
        verbose: bool = False,
    ):
        self.evaluator = ClusterEvaluator(
            X=X,
            cluster_labels=cluster_labels,
            centroids=centroids,
            cluster_names=cluster_names,
            feature_names=feature_names,
            model=model,
            outlier_label=outlier_label,
            metric=metric,
            verbose=verbose,
        )
        self.started_cluster_evaluator = True
        self.evaluator.internal_overview_report()

    def start_cluster_evaluator_from_dataset(
        self, 
        vector_fields:list, 
        alias:str,
        feature_names: Union[list, dict] = None,
        metric: str = "euclidean",
        verbose: bool = False,
        show_progress_bar: bool = False,
    ):
        cluster_field = f"_cluster_.{'.'.join(vector_fields)}.{alias}"
        documents = self.dataset.get_all_documents(
            filters=[
                {
                    "field": field,
                    "filter_type": "exists",
                    "condition": ">=",
                    "condition_value": " ",
                } for field in [cluster_field] + vector_fields
            ],
            select_fields=[cluster_field] + vector_fields,
            show_progress_bar=show_progress_bar,
        )
        self.X = self.dataset.get_field_across_documents(vector_fields[0], documents)
        self.cluster_labels = self.dataset.get_field_across_documents(cluster_field, documents)
        self.centroids = {
            d["_id"] : self.dataset.get_field(vector_fields[0], d)
            for d in self.dataset.datasets.cluster.centroids.documents(
                dataset_id= self.dataset.dataset_id,
                vector_fields=vector_fields,
                alias=alias,
                page_size=9999,
                include_vector=True,
            )["results"]
        }
        self.start_cluster_evaluator(
            self.X, 
            self.cluster_labels, 
            self.centroids,
            # cluster_names=cluster_names,
            feature_names=feature_names,
            # model=model,
            # outlier_label=outlier_label,
            metric=metric,
            verbose=verbose,
        )

    # create wrapper to make sure cluster_evaluator is started

    def section_cluster_report(
        self,
        hierarchy_methods=["ward"],
        color_threshold: float = 1.25,
        plot_method=None,
        add=True,
    ):
        self.h1("Cluster report")
        self.section_cluster_overview_metrics()
        self.section_cluster_dendrogram(
            hierarchy_methods=hierarchy_methods,
            color_threshold=color_threshold,
            plot_method=plot_method,
        )
        self.section_cluster_distance_matrix()
        # self.section_cluster_indepth()

    def section_cluster_overview_metrics(self, add=True):
        self.h2("Overview of cluster metrics")
        self.paragraph(
            "When measuring the performance of a cluster there are 2 main things to look at"
        )
        self.bullet_list(
            [
                "Compactness (intra-cluster distance): Measures the variation within each cluster - how close observations are within each cluster.",
                "Separation (inter-cluster distance): How well separated are the clusters from each other.",
            ]
        )
        for metric, explanation in {
            "davies_bouldin_score" : "      [Compactness, Separation] (0 to infinity, lower is better) This calculates the ratio between each cluster's squared error to the distance between cluster centroids.",
            "calinski_harabasz_score" : "      [Compactness, Separation] (-infinity to infinity, higher is better) Similar to Davies Bouldin score, but also considers the 'group dispersion matrix' that considers the cluster size. Its equivalent to the Variance Ratio Criterion",
            "silhouette_score" : "      [Compactness, Separation] (-1 to 1, higher is better) This is the distance between a sample and all other points in the same cluster, and the same sample to the closest other clusters. This silhouette score is the average of every point’s silhouette score.",
            "total_squared_error_score" : "      [Compactness] (0 to infinity, lower is better) The average squared error between each point of a cluster to its centroid. Its equivalent to inertia.",
        }.items():
            metric_name = " ".join(metric.split("_")).title()
            self.paragraph(
                [self.bold(f"{metric_name}: "), self.evaluator.report[metric]], add=add
            )
            self.paragraph([self.italic(explanation)])
        plot, plotted_method = self.evaluator.plot_boxplot(self.evaluator.report["silhouette_score_summary"], name="Silhouette Score")
        self.plot_by_method(plot, title="Silhouette Score Box Plot", plot_method=plotted_method, add=add)
        plot, plotted_method = self.evaluator.plot_boxplot(self.evaluator.report["squared_error_summary"], name="Squared Error")
        self.plot_by_method(plot, title="Squared Error Box Plot", plot_method=plotted_method, add=add)

    def section_cluster_dendrogram(
        self,
        hierarchy_methods=["ward"],
        color_threshold: float = 1,
        orientation: str = "left",
        plot_method=None,
        add=True,
    ):
        self.h2("Overview of cluster dendrogram")
        self.paragraph(
            "A dendrogram shows the hierarchical relationship between cluster. This can be especially useful to determine which clusters to combined with hierarchical linkage."
        )
        for method in hierarchy_methods:
            height = max(15*self.evaluator.num_clusters, 300)
            plot, plotted_method = self.evaluator.plot_dendrogram(
                hierarchy_method=method,
                plot_method=plot_method,
                color_threshold=color_threshold,
                orientation=orientation,
            )
            self.plot_by_method(
                plot, 
                title=f"{method.title()} linkage dendrogram", 
                plot_method=plotted_method,
                height=height,
                add=add
            )

    def section_cluster_distance_matrix(self, metrics=["cosine", "euclidean"], decimals: int = 4, add=True):
        self.h2("Overview of cluster similarity matrix")
        self.paragraph(
            "Shows a heatmap of the similarity scores between different clusters. This can be especially useful to determine which clusters to combined."
        )
        for metric in metrics:
            plot, plotted_method = self.evaluator.plot_distance_matrix(metric=metric, decimals=decimals)
            chart_title = f"{metric.title()} similarity matrix"
            if metric in ["cosine"]:
                chart_title += " (higher is more similar)"
            else:
                chart_title += " (lower is more similar)"
            self.plot_by_method(plot, title=chart_title, plot_method=plotted_method, add=add)



# -*- coding: utf-8 -*-

from abc import abstractmethod
import pandas as pd
import numpy as np
import json
import gc

from doc_utils import DocUtils

from typing import List, Union, Dict, Any, Tuple, Optional
from typing_extensions import Literal
from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base
from relevanceai.utils.logger import LoguruLogger
from relevanceai.operations.cluster.constants import (
    DIM_REDUCTION,
    DIM_REDUCTION_DEFAULT_ARGS,
)


class DimReductionBase(LoguruLogger, DocUtils):
    def __call__(self, *args, **kwargs):
        return self.fit_predict(*args, **kwargs)

    # @abstractmethod
    def fit_transform(self, *args, **kw) -> np.ndarray:
        raise NotImplementedError

    def fit(self, *args, **kw):
        raise NotImplementedError

    def transform(self, *args, **kw):
        raise NotImplementedError

    def transform_documents(self, vector_fields: List[str], documents: List[Dict]):
        vectors = np.array(
            self.get_fields_across_documents(
                vector_fields, documents, missing_treatment="skip"
            )
        )
        vectors = vectors.reshape(-1, vectors.shape[1] * vectors.shape[2])
        return self.transform(vectors)

    def fit_documents(self, vector_fields: List[str], documents: List[Dict]):
        vectors = np.array(
            self.get_fields_across_documents(
                vector_fields, documents, missing_treatment="skip"
            )
        )
        vectors = vectors.reshape(-1, vectors.shape[1] * vectors.shape[2])
        return self.fit(vectors)

    def get_dr_vector_field_name(self, vector_field: str, alias: str):
        return ".".join(
            [
                "_dr_",
                alias,
                vector_field,
            ]
        )

    def fit_transform_documents(
        self,
        vector_fields: List[str],
        documents: List[Dict],
        alias: str,
        exclude_original_vectors: bool = True,
        dims: int = 3,
    ):
        """
        This function takes a list of documents, a field name, and a dimensionality reduction
        algorithm, and returns a list of documents with a new field containing the dimensionality
        reduced vectors

        Parameters
        ----------
        vector_field : str
            The name of the field in the documents that contains the vectors to be reduced.
        documents : List[Dict]
            The documents to transform.
        alias : str
            The name of the new field that will be created in the documents.
        exclude_original_vectors : bool, optional
            If True, the original vector field will be excluded from the returned documents.
        dims : int, optional
            The number of dimensions to reduce the vectors to.

        Returns
        -------
            A list of documents with the original vector field and the new vector field.

        """

        documents = list(self.filter_docs_for_fields(vector_fields, documents))
        vectors = np.array(
            self.get_fields_across_documents(
                vector_fields, documents, missing_treatment="skip"
            )
        )
        vectors = vectors.reshape(-1, vectors.shape[1] * vectors.shape[2])  # hacky fix
        dr_vectors = self.fit_transform(vectors, dims=dims)
        del vectors  # free more memory, mainly for memory edgecases
        gc.collect()

        if exclude_original_vectors:
            dr_docs = [{"_id": d["_id"]} for d in documents]
            self.set_field_across_documents(alias, dr_vectors, dr_docs)
            return dr_docs
        else:
            self.set_field_across_documents(alias, dr_vectors, documents)
        return documents


# this is mainly for plots
class DimReduction(_Base, DimReductionBase):
    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    @staticmethod
    def dim_reduce(
        vectors: np.ndarray,
        dr: Union[DIM_REDUCTION, DimReductionBase],
        dr_args: Union[None, dict],
        dims: Literal[2, 3],
    ) -> np.ndarray:
        """
        Dimensionality reduction
        """
        if isinstance(dr, str):
            # All relevant DR models
            from relevanceai.operations.dr.models import Ivis, PCA, UMAP, TSNE

            if dr_args is None:
                dr_args = DIM_REDUCTION_DEFAULT_ARGS[dr]
            if dr == "pca":
                return PCA().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            elif dr == "tsne":
                return TSNE().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            elif dr == "umap":
                return UMAP().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            elif dr == "ivis":
                return Ivis().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
            raise ValueError("not suppported")
        elif isinstance(dr, DimReductionBase):
            return dr().fit_transform(vectors=vectors, dr_args=dr_args, dims=dims)
        return np.array([])



from relevanceai.operations_new.transform_base import TransformBase
from relevanceai.constants.errors import MissingPackageError


class KeyWordTransform(TransformBase):
    """
    Extract keyphrase from documents
    """

    def __init__(
        self,
        fields: list,
        model_name: str = "all-mpnet-base-v2",
        lower_bound: int = 0,
        upper_bound: int = 3,
        output_fields: list = None,
        stop_words: list = None,
        max_keywords: int = 1,
        use_maxsum: bool = False,
        nr_candidates: int = 20,
        **kwargs
    ):
        self.fields = fields
        self.model_name = model_name
        self.output_fields = output_fields
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.stop_words = stop_words
        self.max_keywords = max_keywords

        self.use_maxsum = use_maxsum
        self.nr_candidates = nr_candidates
        for k, v in kwargs.items():
            setattr(self, k, v)

    def _get_output_field(self, field):
        return "_keyphrase_." + field

    @property
    def name(self):
        return "keyphrase"

    @property
    def keyphrase_model(self):
        if not hasattr(self, "_model"):
            try:
                from keybert import KeyBERT
            except ModuleNotFoundError:
                raise MissingPackageError("keybert")

            self._model = KeyBERT(self.model_name)
        return self._model

    def extract_keyphrase(self, text):
        keywords = self.keyphrase_model.extract_keywords(
            text,
            keyphrase_ngram_range=(self.lower_bound, self.upper_bound),
            stop_words=self.stop_words,
            top_n=self.max_keywords,
            use_maxsum=self.use_maxsum,
            nr_candidates=self.nr_candidates,
        )
        return [{"keyword": k[0], "score": k[1]} for k in keywords[: self.max_keywords]]

    def extract_keyphrases(self, texts):
        return [self.extract_keyphrase(t) for t in texts]

    def transform(self, documents):
        # Extract the keywords from a bunch of documents
        keyphrase_docs = [{"_id": d["_id"]} for d in documents]
        for i, t in enumerate(self.fields):
            if self.output_fields is not None:
                output_field = self.output_fields[i]
            else:
                output_field = self._get_output_field(t)
            texts = self.get_field_across_documents(t, documents)
            keyphrases = self.extract_keyphrases(texts)
            self.set_field_across_documents(output_field, keyphrases, keyphrase_docs)
        return keyphrase_docs



"""
Base class for ops.py to inherit
"""
from relevanceai.dataset import Dataset
from relevanceai._api import APIClient

from relevanceai.operations_new.ops_run import OperationRun


class OperationsAPILogger:
    pass


class OperationAPIBase(APIClient, OperationRun, OperationsAPILogger):
    @classmethod
    def from_credentials(self, *args, **kwargs):
        """
        .. code-block::

            ClusterOps.from_dataset()
        """
        raise NotImplementedError

    @classmethod
    def from_dataset(self, *args, **kwargs):
        raise NotImplementedError



"""Utilities for Chunk Documents
"""
import itertools
from typing import Callable
from .write_utils import DocWriteUtils


class ChunkDocUtils(DocWriteUtils):
    @classmethod
    def get_chunk(cls, chunk_field, doc):
        return cls.access_document_field(chunk_field, doc)

    @classmethod
    def get_field_across_chunks(cls, chunk_field, field, doc):
        chunk = cls.get_chunk(chunk_field, doc)
        return [cls.get_field(field, d) for d in chunk]

    def run_function_across_chunks(
        self,
        function: Callable,
        chunk_field,
        field: str = None,
        output_field: str = None,
        doc={},
    ):
        """Run a function on a field across chunks
        Params:
        function:
            Any function to run across documents
        field:
            The field is AFTER the Chunk Field
        """
        if field is None:
            chunk_values = self.get_chunk(chunk_field, doc)
        else:
            chunk_values = self.get_field_across_chunks(chunk_field, field, doc)
        if output_field is None:
            return map(function, chunk_values)
        self.set_field_across_documents(
            output_field, list(map(function, chunk_values)), chunk_values
        )

    def run_function_across_chunks_across_docs(
        self,
        function: Callable,
        chunk_field: str,
        field: str = None,
        output_field: str = None,
        docs=[],
    ):
        """Run a function across chunk field in documents"""

        def map_run_function_across_chunks(doc):
            return self.run_function_across_chunks(
                function, chunk_field, field, output_field, doc
            )

        return map(map_run_function_across_chunks, docs)



"""Counting Text operations
"""
from relevanceai.operations_new.processing.text.count.transform import (
    CountTextTransform,
)
from relevanceai.operations_new.ops_base import OperationAPIBase


class CountTextOps(CountTextTransform, OperationAPIBase):
    def __init__(
        self,
        credentials,
        text_fields: list,
        include_char_count: bool = True,
        include_word_count: bool = True,
        include_sentence_count: bool = False,
        output_fields: list = None,
        **kwargs
    ):
        self.text_fields = text_fields
        self.include_char_count = include_char_count
        self.include_word_count = include_word_count
        self.include_sentence_count = include_sentence_count
        self.output_fields = output_fields
        self.credentials = credentials
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def name(self):
        return "count-text"



from typing import Any, List

from relevanceai.operations_new.vectorize.ops import VectorizeOps
from relevanceai.operations_new.vectorize.image.transform import VectorizeImageTransform


class VectorizeImageOps(VectorizeOps, VectorizeImageTransform):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



"""
    Test cleaning the operations
"""


def test_clean_dataset(test_dataset):
    # Simple smoke test to make sure that it runs
    test_dataset.clean_text(text_fields=["sample_1_label"])
    assert True



from copy import deepcopy

from typing import List

from relevanceai.operations_new.transform_base import TransformBase
from relevanceai.utils import MissingPackageError


class ExtractNER(TransformBase):
    def __init__(
        self,
        fields: List[str],
        model_id: str = "dslim/bert-base-NER",
        output_fields: list = None,
        **kwargs
    ):
        self.fields = fields
        self.model_id = model_id
        if output_fields is None:
            self.output_fields = [self._generate_output_field(f) for f in fields]
        else:
            self.output_fields = output_fields  # type: ignore
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def name(self):
        return "ner"

    @property
    def classifier(self):
        if hasattr(self, "_nlp"):
            return self._nlp
        from transformers import AutoTokenizer, AutoModelForTokenClassification
        from transformers import pipeline

        tokenizer = AutoTokenizer.from_pretrained(self.model_id)
        model = AutoModelForTokenClassification.from_pretrained(self.model_id)
        self._nlp = pipeline("ner", model=model, tokenizer=tokenizer)
        return self._nlp

    def extract_ner(
        self,
        text,
    ):
        """The function takes a text and a language as input and returns a list of sentences

        Parameters
        ----------
        text
            The text to split into sentences.
        language : str, optional
            The language of the text.

        Returns
        -------
            A list of sentences.

        """
        entities = self.classifier(text, aggregation_strategy="simple")
        # Sample output looks like this
        # [{'end': 33,
        #   'entity': 'B-ORG',
        #   'index': 8,
        #   'score': 0.9665312767028809,
        #   'start': 30,
        #   'word': 'Hot'},
        #  {'end': 39,
        #   'entity': 'I-ORG',
        #   'index': 9,
        #   'score': 0.9847770929336548,
        #   'start': 34,
        #   'word': 'Wheel'},
        # The is_recorded flag allows us to track whether the
        # entity is just 1 syllable and has been recorded or not
        # Initiate word with something
        for entity in entities:
            entity.pop("score")
            entity.pop("start")
            entity.pop("end")
        return entities

    def extract_ner_from_document(
        self,
        document,
    ):
        """It takes a document, splits it into chunks, and then returns a list of documents, each
        containing a single chunk

        Parameters
        ----------
        fields
            The fields in the document that contain the text to be split.
        document
            The document to be split.
        output_field : str, optional
            The name of the field that will contain the split text.

        Returns
        -------
            A list of dictionaries.

        """
        for i, text_field in enumerate(self.fields):
            text = self.get_field(text_field, document)
            output = self.extract_ner(text)
            self.set_field(
                self.output_fields[i],
                document,
                output,
            )

        return document

    def transform(
        self,
        documents,
    ):
        """It takes a list of documents, and for each document, it splits the text into chunks and adds the
        chunks to the document

        Parameters
        ----------
        fields
            a list of fields in the document that contain text to be split
        documents
            list of documents to be split
        inplace : bool, optional
            bool = True
        output_field : str, optional
            str = "_splittextchunk_"

        Returns
        -------
            A list of documents

        """

        # TODO; switch to something faster than list comprehension
        [
            self.extract_ner_from_document(
                document=document,
            )
            for document in documents
        ]
        return documents



from abc import abstractmethod
from typing import Any, Dict, List

from relevanceai.utils import DocUtils
from tqdm.auto import tqdm


class VectorizeModelBase(DocUtils):
    model_name: str
    has_printed_vector_field_name: dict = {}

    def _get_model_name(self, url):
        model_name = "_".join(url.split("/google/")[-1].split("/")[:-1])
        return model_name

    def vector_name(self, field, output_field: str = None):
        if output_field is not None:
            return output_field
        return f"{field}_{self.model_name}_vector_".replace("/", "_")

    @abstractmethod
    def encode(self, *args, **kwargs):
        pass

    @abstractmethod
    def bulk_encode(self, *args, **kwargs):
        pass

    def print_vector_field_name(self, field: str, output_field: str = None):
        # Store a simple dictionary to test vectorizing
        if field not in self.has_printed_vector_field_name:
            tqdm.write(
                f"Vector field is `{self.vector_name(field, output_field=output_field)}`"
            )
            self.has_printed_vector_field_name[field] = True

    def encode_documents(
        self,
        documents: List[Dict[str, Any]],
        fields: List[str],
        output_fields: List = None,
    ):
        for i, field in enumerate(fields):
            if output_fields is not None:
                self.print_vector_field_name(field, output_fields[i])
            else:
                self.print_vector_field_name(field)
            values = self.get_field_across_documents(field=field, docs=documents)
            vectors = self.bulk_encode(values)
            if output_fields is not None:
                self.set_field_across_documents(
                    field=self.vector_name(field, output_fields[i]),
                    values=vectors,
                    docs=documents,
                )
            else:
                self.set_field_across_documents(
                    field=self.vector_name(field),
                    values=vectors,
                    docs=documents,
                )

        return documents



import pytest

from doc_utils import Document
from doc_utils import DocumentList


def test_document(sample_document):
    document = Document(sample_document)

    label_val = sample_document["value"]

    target_value = document["value"]
    assert label_val == target_value

    document["value"] = 3
    assert document["value"] == 3


def test_nested_document(sample_nested_document):
    document = Document(sample_nested_document)

    label_val = sample_nested_document["value1"]["value2"]["value3"]

    target_value = document["value1.value2.value3"]
    assert label_val == target_value

    document["value1.value2.value3"] = 3
    assert document["value1.value2.value3"] == 3


def test_documents(sample_documents):
    documents = DocumentList(sample_documents)
    assert documents == documents


def test_document_json(sample_documents):
    documents = DocumentList(sample_documents)
    json_docs = documents.json()
    assert all(
        str(sample_document) == str(json_doc)
        for sample_document, json_doc in zip(sample_documents, json_docs)
    )


def test_document_methods(sample_nested_document):
    document = Document(sample_nested_document)

    keys = document.keys()
    values = document.values()
    items = document.items()

    assert keys
    assert values
    assert items



import io
import imageio
import requests

from typing import List, Any, Optional
from urllib.request import Request, urlopen
from skimage import transform

import numpy as np

from relevanceai.operations_new.vectorize.models.base import VectorizeModelBase
from relevanceai.utils.decorators.vectors import catch_errors

try:
    import tensorflow as tf
    import tensorflow_hub as hub
except ModuleNotFoundError as e:
    raise ModuleNotFoundError("Run `pip install tensorflow_hub`.")
except:
    import traceback

    traceback.print_exc()


class TFHubImage2Vec(VectorizeModelBase):
    def __init__(self, url, vector_length, image_dimensions: Optional[int] = None):

        self.model: Any = hub.load(url)
        self.vector_length: int = vector_length
        self.image_dimensions: Optional[int] = image_dimensions
        self.model_name = self._get_model_name(url)

    def read(self, image: Any):
        """It takes an image as a string, bytes, or BytesIO object, and returns a numpy array of the image

        Parameters
        ----------
        image : str
            The image to be read.

        Returns
        -------
            The image is being returned as a numpy array.

        """
        if isinstance(image, str):
            if "http" in image:
                try:
                    b = io.BytesIO(
                        urlopen(
                            Request(image, headers={"User-Agent": "Mozilla/5.0"})
                        ).read()
                    )
                except:
                    return tf.image.decode_jpeg(
                        requests.get(image).content, channels=3, name="jpeg_reader"
                    ).numpy()
            else:
                b = image  # type: ignore
        elif isinstance(image, bytes):
            b = io.BytesIO(image)
        elif isinstance(image, io.BytesIO):
            b = image
        else:
            raise ValueError(
                "Cannot process data type. Ensure it is is string/bytes or BytesIO."
            )
        try:
            return np.array(imageio.imread(b, pilmode="RGB"))
        except:
            return np.array(imageio.imread(b)[:, :, :3])

    def image_resize(
        self, image_array, width=0, height=0, rescale=0, resize_mode="symmetric"
    ):
        if width and height:
            image_array = transform.resize(
                image_array, (width, height), mode=resize_mode, preserve_range=True
            )
        if rescale:
            image_array = transform.rescale(
                image_array, rescale, preserve_range=True, anti_aliasing=True
            )
        return np.array(image_array)

    @catch_errors
    def encode(self, image: str) -> List[float]:
        """It takes an image as input, resizes it to the dimensions specified in the model, and returns the
        output of the model.

        Parameters
        ----------
        image : str
            str

        Returns
        -------
            A list of floats.

        """
        if isinstance(image, str):
            image_tensor = self.read(image)

        if self.image_dimensions is not None:
            if image_tensor.shape[0] != self.image_dimensions:
                raise ValueError(
                    f"Incorrect Image size, please resize to {self.image_dimensions}x{self.image_dimensions}"
                )

        return self.model(image_tensor).numpy().tolist()[0]

    @catch_errors
    def bulk_encode(self, images: List[str]) -> List[List[float]]:
        """The function takes a list of strings and returns a list of lists of floats

        Parameters1
        ----------
        texts : List[str]
            List[str]

        Returns
        -------
            A list of lists of floats.

        """
        images = np.array([self.read(image) for image in images])
        return self.model(images).numpy().tolist()



import io
from relevanceai.apps.report_app.blocks import ReportBlocks


class PyplotReportBlock(ReportBlocks):
    """
    Adding a pyplot block to Report
    """

    def pyplot(
        self,
        fig,
        title: str = "",
        width: int = None,
        height: int = None,
        clear_fig: bool = True,
        dpi: int = 150,
        add: bool = True,
        width_percentage: int = 100,
        **kwargs
    ):
        try:
            import matplotlib.pyplot as plt

            plt.ioff()
        except ImportError:
            raise ImportError(
                ".pyplot requires matplotlib to be installed, install with 'pip install -U matplotlib'."
            )
        if width or height:
            sizes = fig.get_size_inches()
            fig.set_size_inches(
                width / dpi if width else sizes[0], height / dpi if height else sizes[1]
            )
        fig_image = io.BytesIO()
        savefig_kwargs = {"dpi": dpi, "bbox_inches": "tight", "format": "png"}
        savefig_kwargs.update(kwargs)
        fig.savefig(fig_image, **savefig_kwargs)
        self.image(fig_image, title=title, width_percentage=width_percentage, add=add)



import re

from relevanceai._api import APIClient


class Filter(APIClient):
    """
    Filters have been designed to become more pythonic.

    .. code-block::

        old filters = [
            {
                "field": "product_name",
                "filter_type": "exact_match",
                "condition": "==",
                "condition_value": "Durian Leather 2 Seater Sofa"
            }
        ]

        new_filters = dataset["product_name"] == "Durian Leather 2 Seater Sofa"
        # Produces the same as above

        older_filters = [
            {
                "field": "rank",
                "filter_type": "numeric",
                "condition": ">=",
                "condition_value": 2
            },
            {
                "field": "rank",
                "filter_type": "numeric",
                "condition": "<",
                "condition_value": 3
            }
        ]

        new_filters = (dataset["rank"] >= 2) + (dataset["rank"] < 3)

    Exists
    ==============

    Exists filtering can be accessed in a simple way.

    .. code-block::

        old_filters = [
            {
                "field": "brand",
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " "
            }
        ]

        new_filters = dataset["brand"].exists()

        old_filters = [
            {
                "field": "brand",
                "filter_type": "exists",
                "condition": "!=",
                "condition_value": " "
            }
        ]

        new_filters = dataset["brand"].not_exists()

    Contains
    ==============

    Same with contains.

    .. code-block::

        old_filters = [
            {
                "field": "description",
                "filter_type": "contains",
                "condition": "==",
                "condition_value": "Durian BID"
            }
        ]

        new_filters = dataset["description"].contains("Durian BID")

    Dates
    ==============

    Date filtering

    .. code-block::

        old_filters = [
            {
                "field": ""insert_date_"",
                "filter_type": "date",
                "condition": "==",
                "condition_value": "2020-07-01"
            }
        ]

        new_filters = dataset["_insert_date"].date(2020-07-01")

    """

    def __init__(self, field, dataset_id, condition, condition_value, **kwargs):
        super().__init__(kwargs["credentials"])
        kwargs.pop("credentials")

        self.field = field
        self.dataset_id = dataset_id
        self.condition = condition
        self.condition_value = condition_value

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def dtype(self):
        schema = self.datasets.schema(self.dataset_id)
        return schema[self.field]

    def get(self):
        # get dtype in elif statements to avoid
        # schema errors if field does not exist yet
        if hasattr(self, "filter_type"):
            filter_type = self.filter_type

        elif self.dtype == "numeric":
            filter_type = "numeric"

        elif self.dtype == "date":
            filter_type = "date"

        else:
            filter_type = "exact_match"

        return [
            {
                "field": self.field,
                "filter_type": filter_type,
                "condition": self.condition,
                "condition_value": self.condition_value,
            }
        ]



"""
Clean HTML
"""
import string
import re
import warnings

from typing import List, Optional
from collections import Counter
from html.parser import HTMLParser
from io import StringIO


class BaseTextProcessing:
    """Base text processing"""

    @staticmethod
    def remove_punctuation(text):
        return "".join([ch for ch in text if ch not in string.punctuation])

    @staticmethod
    def lower_text(text):
        return text.lower()

    @staticmethod
    def remove_html_tags(text):
        stripper = MLStripper()
        return stripper.clean(text)

    @staticmethod
    def remove_digits(text):
        return "".join([ch for ch in text if ch not in string.digits])

    @staticmethod
    def remove_stopwords(text: str, additional_stp_wrds: List[str] = None):
        import nltk
        from nltk.corpus import stopwords
        from nltk.tokenize import word_tokenize

        nltk.download("punkt")
        nltk.download("wordnet")
        nltk.download("omw-1.4")
        nltk.download("stopwords")
        stop_words = set(stopwords.words("english"))
        if additional_stp_wrds:
            stop_words = stop_words.union(set([w.lower() for w in additional_stp_wrds]))
        word_tokens = word_tokenize(text)
        return " ".join([w for w in word_tokens if w.lower() not in stop_words])

    @staticmethod
    def remove_url(text):
        return re.sub(
            "http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+",
            "",
            text,
            flags=re.MULTILINE,
        )

    @staticmethod
    def lemmatize(text: str):
        import nltk

        nltk.download("punkt")
        nltk.download("omw-1.4")
        nltk.download("wordnet")
        from nltk.stem import WordNetLemmatizer
        from nltk.tokenize import word_tokenize

        # todo: find a better one => (NLTK changes less to le !!!)
        wordnet_lemmatizer = WordNetLemmatizer()
        word_tokens = word_tokenize(text)
        return " ".join([wordnet_lemmatizer.lemmatize(w) for w in word_tokens])

    @staticmethod
    def replace_words(text, replace_words: dict):
        for old, new in replace_words.items():
            if old in text:
                text = text.replace(old, new)
        return text

    @staticmethod
    def get_word_frequency(
        str_list: List[str],
        remove_stop_words: bool = True,
        additional_stop_words: Optional[List[str]] = None,
        language="english",
    ) -> List:
        """Returns a sorted word frequency in Python"""
        additional_stop_words = (
            [] if additional_stop_words is None else additional_stop_words
        )
        try:
            import nltk

            nltk.download("stopwords")
            from nltk.corpus import stopwords
        except ModuleNotFoundError:
            warnings.warn("You are missing NLTK, please run `pip install nltk`")

        if remove_stop_words:
            stpw = stopwords.words(language)
            stpw += additional_stop_words
        else:
            stpw = []
        word_counter = Counter(
            [w.lower() for s in str_list for w in s.split() if w not in stpw]
        )
        return sorted(word_counter.items(), key=lambda item: (-item[1], item[0]))


class MLStripper(HTMLParser):
    """Remove HTML from the code and retrieves data."""

    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def handle_data(self, d):
        self.text.write(d)

    def get_data(self):
        return (
            self.text.getvalue()
            .replace("\r", "")
            .replace("\n", "")
            .replace("\t", " ")
            .strip()
        )

    def clean(self, text):
        self.reset()
        self.handle_data(text)
        return self.get_data()



from typing import Any

from relevanceai.operations_new.vectorize.transform import VectorizeTransform
from relevanceai.operations_new.vectorize.models.base import VectorizeModelBase
from relevanceai.operations_new.vectorize.models.text.mappings import *


class VectorizeTextTransform(VectorizeTransform):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _get_model(self, model: Any) -> VectorizeModelBase:
        """If the model is a string, then it is either a TFHUB model or a Sentence Transformer model. If it
        is a TFHUB model, then return None. If it is a Sentence Transformer model, then return a
        SentenceTransformer2Vec object

        Parameters
        ----------
        model : Any
            str = "bert-base-nli-mean-tokens"

        Returns
        -------
            The model is being returned.

        """
        if isinstance(model, str):
            if model in TFHUB_MODELS:
                from relevanceai.operations_new.vectorize.models.text.tfhub import (
                    TFHubText2Vec,
                )

                vector_length = TFHUB_MODELS[model]["vector_length"]

                model = TFHubText2Vec(
                    url=model,
                    vector_length=vector_length,
                )

                return model

            elif model in SENTENCE_TRANSFORMER_MODELS:
                from relevanceai.operations_new.vectorize.models.text.sentence_transformers import (
                    SentenceTransformer2Vec,
                )
                from sentence_transformers import SentenceTransformer

                vector_length = SENTENCE_TRANSFORMER_MODELS[model]["vector_length"]

                model = SentenceTransformer2Vec(
                    model=SentenceTransformer(model),
                    vector_length=vector_length,
                    model_name=model,
                )

                return model

            else:  # assume model is sentence transformer model
                from relevanceai.operations_new.vectorize.models.text.sentence_transformers import (
                    SentenceTransformer2Vec,
                )
                from sentence_transformers import SentenceTransformer

                vector_length = None

                model = SentenceTransformer2Vec(
                    model=SentenceTransformer(model),
                    vector_length=vector_length,
                    model_name=model,
                )

                return model

        elif isinstance(model, VectorizeModelBase):
            return model

        else:
            raise ValueError(
                "Model should either be a supported model string or inherit from ModelBase"
            )



from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base
from relevanceai._api.endpoints.datasets.cluster.centroids import CentroidsClient
from typing import Optional, List


class ClusterClient(_Base):
    def __init__(self, credentials: Credentials):
        self.centroids = CentroidsClient(credentials)
        super().__init__(credentials)

    def aggregate(
        self,
        dataset_id: str,
        vector_fields: list,
        metrics: Optional[list] = None,
        groupby: Optional[list] = None,
        sort: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        alias: str = "default",
    ):
        """
        Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.
        It can only can be used after a vector field has been clustered. \n
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        vector_fields : list
            The vector field that was clustered on
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten
        alias: string
            Alias used to name a vector field. Belongs in field_{alias}vector
        """
        metrics = [] if metrics is None else metrics
        groupby = [] if groupby is None else groupby
        sort = [] if sort is None else sort
        filters = [] if filters is None else filters

        endpoint = f"/datasets/{dataset_id}/cluster/aggregate"
        method = "POST"
        parameters = {
            "dataset_id": dataset_id,
            "aggregation_query": {"groupby": groupby, "metrics": metrics, "sort": sort},
            "filters": filters,
            "page_size": page_size,
            "page": page,
            "asc": asc,
            "flatten": flatten,
            "vector_fields": vector_fields,
            "alias": alias,
        }
        # self._log_to_dashboard(
        #     method=method,
        #     parameters=parameters,
        #     endpoint=endpoint,
        #     dashboard_type="cluster_aggregation",
        # )
        return self.make_http_request(
            endpoint=endpoint, method=method, parameters=parameters
        )

    def merge(
        self,
        dataset_id: str,
        vector_fields: List[str],
        alias: str,
        cluster_ids: list,
        centroid_vector_fields: List[str] = None,
    ):
        """
        Merge clusters together
        """
        endpoint = f"/datasets/{dataset_id}/cluster/merge"
        method = "POST"
        if centroid_vector_fields is None:
            centroid_vector_fields = []
        parameters = {
            "dataset_id": dataset_id,
            "vector_fields": vector_fields,
            "centroid_vector_fields": centroid_vector_fields,
            "alias": alias,
            "cluster_ids": cluster_ids,
        }
        return self.make_http_request(
            endpoint=endpoint, method=method, parameters=parameters
        )



"""Multithreading Module
"""
import math

import threading
import multiprocessing as mp

from tqdm.auto import tqdm
from typing import Any, Callable, Dict, List, Optional, Tuple

from concurrent.futures import (
    as_completed,
    wait,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
)
from relevanceai.utils.json_encoder import json_encoder
from relevanceai.utils.progress_bar import progress_bar


def chunk(iterables, n=20):
    return [iterables[i : i + n] for i in range(0, int(len(iterables)), int(n))]


def multithread(
    func, iterables, max_workers=2, chunksize=20, show_progress_bar: bool = False
):
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        progress_tracker = progress_bar(
            range(math.ceil(len(iterables) / chunksize)),
            show_progress_bar=show_progress_bar,
        )

        futures = [executor.submit(func, it) for it in chunk(iterables, chunksize)]

        if show_progress_bar:
            with progress_tracker as pt:
                for _ in as_completed(futures):
                    if hasattr(pt, "update"):
                        pt.update(1)
        else:
            wait(futures)

        return [future.result() for future in futures]


def multiprocess(
    func,
    iterables,
    max_workers=2,
    chunksize=20,
    post_func_hook: Callable = None,
    show_progress_bar: bool = False,
    process_args: tuple = (),
):
    # with progress_bar(total=int(len(iterables) / chunksize),
    #     show_progress_bar=show_progress_bar) as pbar:
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        # Create trackers
        progress_tracker = progress_bar(
            range(math.ceil(len(iterables) / chunksize)),
            show_progress_bar=show_progress_bar,
        )
        progress_iterator = iter(progress_tracker)

        if len(process_args) > 0:
            futures = [
                executor.submit(func, it, process_args)
                for it in chunk(iterables, chunksize)
            ]
        else:
            futures = [executor.submit(func, it) for it in chunk(iterables, chunksize)]
        results = []
        for future in as_completed(futures):
            if post_func_hook:
                results.append(post_func_hook(future.result()))
            else:
                results.append(future.result())
            if progress_tracker is not None:
                next(progress_iterator)
            if show_progress_bar is True:
                progress_tracker.update(1)
        return results


def multiprocess_list(
    func,
    iterables,
    max_workers=2,
    chunksize=1,
    post_func_hook: Callable = None,
    show_progress_bar: bool = False,
    process_args: tuple = (),
):
    # with progress_bar(total=int(len(iterables) / chunksize),
    #     show_progress_bar=show_progress_bar) as pbar:
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        # Create trackers
        progress_tracker = progress_bar(
            range(math.ceil(len(iterables) / chunksize)),
            show_progress_bar=show_progress_bar,
        )
        progress_iterator = iter(progress_tracker)

        if len(process_args) > 0:
            futures = [executor.submit(func, it, process_args) for it in iterables]
        else:
            futures = [executor.submit(func, it) for it in iterables]
        results = []
        for future in as_completed(futures):
            if post_func_hook:
                results.append(post_func_hook(future.result()))
            else:
                results.append(future.result())
            if progress_tracker is not None:
                next(progress_iterator)
            if show_progress_bar is True:
                progress_tracker.update(1)
        return results


class Push:
    def __init__(
        self,
        dataset,
        func: Callable,
        documents: List[Dict[str, Any]],
        func_kwargs: Dict[str, Any],
        chunksize: Optional[int] = None,
        max_workers: Optional[int] = None,
        background_execution: bool = False,
    ):
        from relevanceai.dataset.dataset import Dataset

        self.dataset: Dataset = dataset
        self.dataset_id: str = dataset.dataset_id

        self.func = func
        self.func_kwargs = func_kwargs

        self.frontier = {document["_id"]: 0 for document in documents}
        self.push_queue: mp.Queue = mp.Queue(maxsize=len(documents))
        for document in documents:
            self.push_queue.put(document)

        self.chunksize = chunksize
        self.max_workers = 2 if max_workers is None else max_workers

        self.func_kwargs["return_documents"] = True
        show_progress_bar = self.func_kwargs.pop("show_progress_bar", True)

        self.lock = threading.Lock()
        self.tqdm_kwargs = dict(leave=True, disable=(not show_progress_bar))
        self.insert_count = 0
        self.background_execution = background_execution

        self.push_bar = tqdm(
            range(len(documents)),
            desc="push",
            **self.tqdm_kwargs,
        )

    @property
    def failed_ids(self) -> List[str]:
        return [document for document, fails in self.frontier.items() if fails > 0]

    def _get_batch(self):
        batch = []
        while True:
            if len(batch) >= self.chunksize:
                break
            # if self.push_queue.empty():
            #     break
            try:
                document = self.push_queue.get(timeout=1)
            except:
                break
            batch.append(document)
        return batch

    def _handle_failed_documents(
        self,
        res: Dict[str, Any],
        batch: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:

        failed_documents = res["response_json"]["failed_documents"]
        readded_documents = []

        if failed_documents:
            with self.lock:
                desc = f"push - failed_documents = {len(self.failed_ids)}"
                self.push_bar.set_description(desc)

            # ...find these failed documents within the batch...
            failed_ids = set(map(lambda x: x["_id"], failed_documents))
            failed_documents = [
                document for document in batch if document["_id"] in failed_ids
            ]

            # ...and re add them to the push queue
            for failed_document in failed_documents:
                _id = failed_document["_id"]
                if self.frontier[_id] <= 3:
                    self.frontier[_id] += 1
                    self.push_queue.put(failed_document)
                    readded_documents.append(failed_document)

        return readded_documents

    def _push(self) -> None:

        while True:
            with self.lock:
                batch = self._get_batch()

            if not batch:
                break

            batch = json_encoder(batch)
            result = self.func(
                dataset_id=self.dataset_id, documents=batch, **self.func_kwargs
            )

            failed_documents = self._handle_failed_documents(result, batch)

            inserted = len(batch) - len(failed_documents)

            with self.lock:
                self.insert_count += inserted
                self.push_bar.update(inserted)

    def run(self) -> Tuple[int, List[str]]:
        push_threads = [
            threading.Thread(target=self._push) for _ in range(self.max_workers)
        ]
        for thread in push_threads:
            thread.start()

        if not self.background_execution:
            for thread in push_threads:
                thread.join()

        return self.insert_count, self.failed_ids
